"""Wrappers for DPF operators.

These operators are available as functions from ``dpf.operators`` and
simplify the creation of chained operators.
"""
from ansys import dpf
from ansys.dpf.core.common import types as dpf_types


def _check_type(instance, allowable_type):
    if not isinstance(instance, allowable_type):
        if isinstance(allowable_type, tuple):
            names = ", ".join([atype.__name__ for atype in allowable_type])
            raise TypeError("Input type must be one of (%s)" % names)
        else:
            raise TypeError("Input type must be a %s" % allowable_type.__name__)


# TODO: deprecate this file
def sum(var_inp):
    """Sum all elementary data of a field to get one elementary data.

    If an operator, it must contain only one field.

    Parameters
    ----------
    var_inp : ansys.dpf.core.FieldsContainer, ansys.dpf.core.Field, or ansys.dpf.core.Operator
        Field or fields container containing only one field.

    Returns
    -------
    sum : Field
        Sum of the input.
    """
    # Examples
    # --------
    # Sum element volume using a model.

    # >>> from ansys.dpf import core as dpf
    # >>> from ansys.dpf.core import examples
    # >>> model = dpf.Model(examples.static_rst)
    # >>> e_vol = model.results.volume()
    # >>> total_volume = e_vol.sum()

    # Sum element volume using dpf.core.sum.

    # >>> from ansys.dpf.core import examples
    # >>> model = dpf.Model(examples.static_rst)
    # >>> e_vol = model.results.volume()
    # >>> total_volume = dpf.help.sum(e_vol)

    typ_err = TypeError(
        "Input must be a Field or FieldsContainer containing "
        "only one field, or an operator that returns one Field"
    )

    # Create an accumulate operator.
    sum_op = dpf.core.Operator("accumulate")

    if isinstance(var_inp, dpf.core.Field):
        sum_op.connect(var_inp)
    elif isinstance(var_inp, dpf.core.FieldsContainer):
        if len(var_inp) > 1:
            raise typ_err
        sum_op.connect(var_inp)
    elif isinstance(var_inp, dpf.core.Operator):
        return _sum_oper(var_inp)
    else:
        raise typ_err

    return sum_op.get_output(0, dpf_types.field)


def _sum_oper(oper):
    """Sum all elementary data of the operator.

    An operator must contain only one field.

    Returns
    -------
    sum : Field
        Sum of this operator.

    """
    fields = oper.fields
    if len(fields) > 1:
        raise TypeError(
            "Input must be a Field or FieldsContainer containing "
            "only one field, or an operator that returns one Field"
        )

    sum_op = dpf.core.Operator("accumulate")

    sum_op.connect(0, fields)
    field = sum_op.get_output(0, dpf_types.field)

    if oper.physics_name:
        field._name = f"Sum of {field.physics_name}"
    else:
        field._name = f"Sum"
    field._unit = field._unit
    return field


def to_nodal(var_inp):
    """Transform the elemental field into a nodal field using an averaging process.

    Returns
    -------
    field : Field
        Field containing the nodal averaged field.
    """
    # try to use the same server as input
    _check_type(var_inp, dpf.core.Field)
    if isinstance(var_inp, dpf.core.Field):
        oper = dpf.core.Operator("to_nodal")
        return oper.get_output(0, dpf_types.field)
    elif isinstance(var_inp, dpf.core.FieldsContainer):
        oper = dpf.core.Operator("to_nodal_fc")
        return oper.get_output(0, dpf_types.fields_container)
    else:
        raise TypeError("Input type must be a Field, FieldContainer")


def norm(var_inp):
    """Retrieve the Euclidean norm of a field, field container, or operator.

    Returns
    -------
    field : Field, ansys.dpf.core.FieldContainer, or ansys.dpf.core.Operator
        Euclidean norm of this field. The output type will match the input type.
    """
    if isinstance(var_inp, dpf.core.Field):
        return _norm(var_inp)
    elif isinstance(var_inp, dpf.core.FieldsContainer):
        return _norm_fc(var_inp)
    elif isinstance(var_inp, dpf.core.Operator):
        return _norm_op(var_inp)
    else:
        raise TypeError("Input type must be a Field, FieldContainer, or an Operator")


def _norm(field):
    """Retrieve the Euclidean norm of a field.

    Returns
    -------
    field : ansys.dpf.core.Field
        Euclidean norm of this field.
    """
    _check_type(field, dpf.core.Field)
    norm_op = dpf.core.Operator("norm")
    norm_op.connect(0, field)
    norm_field = norm_op.get_output(0, dpf_types.field)
    return norm_field


def _norm_fc(fields):
    """Retrieve the Euclidean norm for the elementary data fields.

    Returns
    -------
    fields : ansys.dpf.core.FieldsContainer
        Euclidean norm of the fields.
    """
    _check_type(fields, dpf.core.FieldsContainer)
    norm_op = dpf.core.Operator("norm_fc")
    norm_op.connect(0, fields)
    norm_fields = norm_op.get_output(0, dpf_types.fields_container)
    return norm_fields


def _norm_op(oper):
    """Retrieve a chained norm operator.
    Returns
    -------
    oper : ansys.dpf.core.Operator
        Chained Euclidean norm operator.
    """
    _check_type(oper, dpf.core.Operator)

    # try to use the same server as input
    norm_op = dpf.core.Operator("norm_fc")
    norm_op.inputs.connect(oper.outputs)
    return norm_op


def eqv(var_inp):
    """Retrieve the Von Mises stress of a field or field container.

    Returns
    -------
    field : ansys.dpf.core.Field, ansys.dpf.core.FieldContainer
        The von Mises stress of this field. The output type will match the input type.
    """
    if isinstance(var_inp, dpf.core.Field):
        return _eqv(var_inp)
    elif isinstance(var_inp, dpf.core.FieldsContainer):
        return _eqv_fc(var_inp)
    # elif isinstance(var_inp, dpf.core.Operator):
    # return _eqv_op(var_inp)
    else:
        raise TypeError("Input type must be a Field or FieldContainer")


# TODO: Consider combining eqv and eqv_fc
def _eqv(field):
    """Retrieve the von Mises stress field.

    Parameters
    ----------
    field : ansys.dpf.core.Field
        Field containing the component stresses.

    Returns
    -------
    field : ansys.dpf.core.Field
        Field containing the von Mises stress.
    """
    _check_type(field, dpf.core.Field)
    oper = dpf.core.operator("eqv")
    oper.connect(0, field)
    field = oper.get_output(0, dpf_types.field)
    return field


def _eqv_fc(fields):
    """Compute the element-wise von Mises criteria for each tensor in the fields of the field container.

    Returns
    -------
    fields : ansys.dpf.core.FieldsContainer
        Element-wise von Mises criteria for this field container.
    """
    _check_type(fields, dpf.core.FieldsContainer)
    oper = fields._model.operator("eqv_fc")
    oper.connect(0, fields)
    eqv_fields = oper.get_output(0, dpf_types.fields_container)
    return eqv_fields


def min_max(var_inp):
    """Retrieve the minimum/maximum operator for a field, fields container, or operator input.

    Returns
    -------
    oper : ansys.dpf.core.Operator
        Component-wise minimum/maximum operator over the input.
    """
    if isinstance(var_inp, dpf.core.Field):
        return _min_max(var_inp)
    elif isinstance(var_inp, dpf.core.FieldsContainer):
        return _min_max_fc(var_inp)
    elif isinstance(var_inp, dpf.core.Operator):
        return _min_max_oper(var_inp)
    else:
        raise TypeError("Input type must be a Field or FieldContainer")


def _min_max(field):
    """Retrieve the minimum/maximum operator for the field.

    Returns
    -------
    oper : ansys.dpf.core.Operator
        Component-wise minimum/maximum operator over the input.
    """

    oper = dpf.core.Operator("min_max")
    oper.inputs.connect(field)
    return oper


def _min_max_fc(fields):
    """Retrieve the minimum/maximum operator for the field container.

    Returns
    -------
    oper : ansys.dpf.core.Operator
        Component-wise minimum/maximum operator over a field
        container.
    """
    oper = dpf.core.Operator("min_max_fc")
    oper.connect(0, fields)
    return oper


def _min_max_oper(oper):
    """Retrieve a chained minimum/maximum operator.

    Returns
    -------
    oper : ansys.dpf.core.Operator
        Component-wise minimum/maximum operator.
    """

    min_max_oper = dpf.core.Operator("min_max_fc")
    min_max_oper.connect(0, oper, 0)
    return min_max_oper


def add(a, b):
    """Add two fields.

    Parameters
    ----------
    a : ansys.dpf.core.Field or ansys.dpf.core.FieldContainer
        Field or fields container with only one field.
    b : ansys.dpf.core.Field or ansys.dpf.core.FieldContainer
        Field or fields container with only one field.

    Returns
    -------
    field_sum : ansys.dpf.core.Field
        Sum of the two fields.
    """
    _check_type(a, (dpf.core.Field, dpf.core.FieldsContainer))
    _check_type(b, (dpf.core.Field, dpf.core.FieldsContainer))

    sum_oper = dpf.core.Operator("add")
    sum_oper.connect(0, a)
    sum_oper.connect(1, a)
    return sum_oper.get_output(0, dpf.core.types.field)


def element_dot(a, b):
    """Compute the element-wise dot product between two vector fields.

    Parameters
    ----------
    a : ansys.dpf.core.Field or ansys.dpf.core.FieldContainer
        Field or fields container with only one field.

    b : ansys.dpf.core.Field or ansys.dpf.core.FieldContainer
        Field or fields container with only one field.

    Returns
    -------
    field_sum : ansys.dpf.core.Field
        Element-wise dot product of the two fields.

    Examples
    --------
    Compute the element-wise dot product.

    >>> from ansys.dpf import core as dpf
    >>> import numpy as np
    >>> data = np.random.random((10, 3))
    >>> field_a = dpf.field_from_array(data)
    >>> field_b = dpf.field_from_array(data)
    >>> fout = dpf.help.element_dot(field_a, field_b)
    >>> fout.shape
    10

    >>> # Numpy equivalent
    >>> arr_a = np.random.random((10, 3))
    >>> arr_b = np.random.random((10, 3))
    >>> edot = np.sum(arr_a*arr_b, 1)
    >>> edot.shape
    (10,)

    """
    _check_type(a, (dpf.core.Field, dpf.core.FieldsContainer))
    _check_type(b, (dpf.core.Field, dpf.core.FieldsContainer))

    op = dpf.core.Operator("dot")
    op.connect(0, a)
    op.connect(1, b)
    return op.get_output(0, dpf.core.types.field)


def sqr(field):
    """Compute the element-wise square of a field.

    Parameters
    ----------
    field : ansys.dpf.core.Field or ansys.dpf.core.FieldContainer
        Field or fields container with only one field.

    Returns
    -------
    field_sqr : ansys.dpf.core.Field
        Element-wise square of the field.

    Examples
    --------
    Use the built-in operator.

    >>> from ansys.dpf import core as dpf
    >>> field = dpf.field_from_array([1, 8])
    >>> field_sqr = field**2
    >>> field_sqr.outputs.field().data
    array([ 1., 64.])

    Use the operator method.

    >>> from ansys.dpf import core as dpf
    >>> field = dpf.field_from_array([1, 8])
    >>> field_sqr = dpf.help.sqr(field)
    >>> print(field_sqr.data)
    [ 1. 64.]

    """

    _check_type(field, (dpf.core.Field, dpf.core.FieldsContainer))
    op = dpf.core.Operator("sqr")
    op.connect(0, field)
    return op.get_output(0, dpf.core.types.field)


def dot_tensor(a, b):
    """Compute the element-wise dot product between two tensor fields.

    Parameters
    ----------
    a : ansys.dpf.core.Field or ansys.dpf.core.FieldContainer
        Field or fields container with only one field.
    b : ansys.dpf.core.Field or ansys.dpf.core.FieldContainer
        Field or fields container with only one field.

    Returns
    -------
    field_sum : ansys.dpf.core.Field
        Element-wise dot product between two vector fields.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> import numpy as np
    >>> arr_a = np.ones((5, 3))
    >>> arr_a[:, 2] = 0
    >>> arr_b = np.ones((5, 3))
    >>> arr_b[:, 1] = 0
    >>> field_a = dpf.field_from_array(arr_a)
    >>> field_b = dpf.field_from_array(arr_b)
    >>> field_out = dpf.help.dot_tensor(field_a, field_b)
    >>> field_out.data
    array([[1., 1., 0., 0., 0., 0., 1., 1., 0.],
           [1., 1., 0., 0., 0., 0., 1., 1., 0.],
           [1., 1., 0., 0., 0., 0., 1., 1., 0.],
           [1., 1., 0., 0., 0., 0., 1., 1., 0.],
           [1., 1., 0., 0., 0., 0., 1., 1., 0.]])

    """
    _check_type(a, (dpf.core.Field, dpf.core.FieldsContainer))
    _check_type(b, (dpf.core.Field, dpf.core.FieldsContainer))

    op = dpf.core.Operator("dot_tensor")
    op.connect(0, a)
    op.connect(1, b)
    return op.get_output(0, dpf.core.types.field)



import socket


def get_local_ip():
    """Retrieve the local IP address to connect to the gRPC service.

    Returns
    -------
    str
        Local IP address to connect with the gRPC service.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        try:
            # doesn't even have to be reachable
            s.connect(("10.255.255.255", 1))
        except socket.error:
            return "127.0.0.1"

        return s.getsockname()[0]


ip = get_local_ip()
port = 50052
environment = "windows"
configuration = "release"



from .plastic_strain_principal_1 import plastic_strain_principal_1 
from .plastic_strain_principal_3 import plastic_strain_principal_3 
from .rigid_transformation import rigid_transformation 
from .elastic_strain_Y import elastic_strain_Y 
from .nodal_moment import nodal_moment 
from .elemental_mass import elemental_mass 
from .heat_flux import heat_flux 
from .co_energy import co_energy 
from .plastic_strain_principal_2 import plastic_strain_principal_2 
from .elastic_strain_Z import elastic_strain_Z 
from .stress import stress 
from .stress_X import stress_X 
from .stress_Y import stress_Y 
from .stress_Z import stress_Z 
from .stress_XY import stress_XY 
from .stress_YZ import stress_YZ 
from .modal_basis import modal_basis 
from .stress_XZ import stress_XZ 
from .stress_principal_1 import stress_principal_1 
from .stress_principal_2 import stress_principal_2 
from .stress_principal_3 import stress_principal_3 
from .elastic_strain import elastic_strain 
from .elastic_strain_X import elastic_strain_X 
from .elastic_strain_XY import elastic_strain_XY 
from .elastic_strain_YZ import elastic_strain_YZ 
from .elastic_strain_XZ import elastic_strain_XZ 
from .elastic_strain_principal_1 import elastic_strain_principal_1 
from .elastic_strain_principal_2 import elastic_strain_principal_2 
from .elastic_strain_principal_3 import elastic_strain_principal_3 
from .plastic_strain import plastic_strain 
from .plastic_strain_X import plastic_strain_X 
from .plastic_strain_Y import plastic_strain_Y 
from .plastic_strain_Z import plastic_strain_Z 
from .hydrostatic_pressure import hydrostatic_pressure 
from .plastic_strain_XY import plastic_strain_XY 
from .plastic_strain_YZ import plastic_strain_YZ 
from .plastic_strain_XZ import plastic_strain_XZ 
from .acceleration import acceleration 
from .acceleration_X import acceleration_X 
from .acceleration_Y import acceleration_Y 
from .element_centroids import element_centroids 
from .acceleration_Z import acceleration_Z 
from .reaction_force import reaction_force 
from .velocity import velocity 
from .velocity_X import velocity_X 
from .velocity_Y import velocity_Y 
from .velocity_Z import velocity_Z 
from .displacement import displacement 
from .displacement_X import displacement_X 
from .displacement_Y import displacement_Y 
from .displacement_Z import displacement_Z 
from .heat_flux_X import heat_flux_X 
from .electric_field import electric_field 
from .heat_flux_Y import heat_flux_Y 
from .heat_flux_Z import heat_flux_Z 
from .element_nodal_forces import element_nodal_forces 
from .structural_temperature import structural_temperature 
from .incremental_energy import incremental_energy 
from .stiffness_matrix_energy import stiffness_matrix_energy 
from .thermal_strain import thermal_strain 
from .eqv_stress_parameter import eqv_stress_parameter 
from .stress_ratio import stress_ratio 
from .accu_eqv_plastic_strain import accu_eqv_plastic_strain 
from .plastic_state_variable import plastic_state_variable 
from .accu_eqv_creep_strain import accu_eqv_creep_strain 
from .plastic_strain_energy_density import plastic_strain_energy_density 
from .material_property_of_element import material_property_of_element 
from .creep_strain_energy_density import creep_strain_energy_density 
from .elastic_strain_energy_density import elastic_strain_energy_density 
from .contact_status import contact_status 
from .contact_penetration import contact_penetration 
from .contact_pressure import contact_pressure 
from .contact_friction_stress import contact_friction_stress 
from .contact_total_stress import contact_total_stress 
from .contact_sliding_distance import contact_sliding_distance 
from .contact_gap_distance import contact_gap_distance 
from .contact_surface_heat_flux import contact_surface_heat_flux 
from .num_surface_status_changes import num_surface_status_changes 
from .contact_fluid_penetration_pressure import contact_fluid_penetration_pressure 
from .elemental_volume import elemental_volume 
from .artificial_hourglass_energy import artificial_hourglass_energy 
from .kinetic_energy import kinetic_energy 
from .thermal_dissipation_energy import thermal_dissipation_energy 
from .nodal_force import nodal_force 
from .temperature import temperature 
from .raw_displacement import raw_displacement 
from .raw_reaction_force import raw_reaction_force 
from .electric_potential import electric_potential 
from .thickness import thickness 
from .element_orientations import element_orientations 
from .custom import custom 
from .stress_von_mises import stress_von_mises 
from .cyclic_expansion import cyclic_expansion 
from .equivalent_radiated_power import equivalent_radiated_power 
from .torque import torque 
from .cyclic_analytic_usum_max import cyclic_analytic_usum_max 
from .cyclic_analytic_seqv_max import cyclic_analytic_seqv_max 
from .recombine_harmonic_indeces_cyclic import recombine_harmonic_indeces_cyclic 
from .poynting_vector import poynting_vector 
from .poynting_vector_surface import poynting_vector_surface 
from .nodal_averaged_elastic_strains import nodal_averaged_elastic_strains 
from .add_rigid_body_motion import add_rigid_body_motion 
from .nodal_averaged_equivalent_elastic_strain import nodal_averaged_equivalent_elastic_strain 
from .run import run 
from .cyclic_expanded_velocity import cyclic_expanded_velocity 
from .cyclic_expanded_el_strain import cyclic_expanded_el_strain 
from .nodal_averaged_thermal_swelling_strains import nodal_averaged_thermal_swelling_strains 
from .nodal_averaged_stresses import nodal_averaged_stresses 
from .nodal_averaged_thermal_strains import nodal_averaged_thermal_strains 
from .nodal_averaged_plastic_strains import nodal_averaged_plastic_strains 
from .nodal_averaged_creep_strains import nodal_averaged_creep_strains 
from .nodal_averaged_equivalent_thermal_strains import nodal_averaged_equivalent_thermal_strains 
from .nodal_averaged_equivalent_plastic_strain import nodal_averaged_equivalent_plastic_strain 
from .nodal_averaged_equivalent_creep_strain import nodal_averaged_equivalent_creep_strain 
from .euler_nodes import euler_nodes 
from .nmisc import nmisc 
from .enf_rotation_by_euler_nodes import enf_rotation_by_euler_nodes 
from .cms_matrices_provider import cms_matrices_provider 
from .smisc import smisc 
from .nodal_rotation_by_euler_nodes import nodal_rotation_by_euler_nodes 
from .stress_rotation_by_euler_nodes import stress_rotation_by_euler_nodes 
from .elastic_strain_rotation_by_euler_nodes import elastic_strain_rotation_by_euler_nodes 
from .plastic_strain_rotation_by_euler_nodes import plastic_strain_rotation_by_euler_nodes 
from .pres_to_field import pres_to_field 
from .prns_to_field import prns_to_field 
from .remove_rigid_body_motion import remove_rigid_body_motion 
from .remove_rigid_body_motion_fc import remove_rigid_body_motion_fc 
from .add_rigid_body_motion_fc import add_rigid_body_motion_fc 
from .cyclic_expanded_displacement import cyclic_expanded_displacement 
from .cyclic_expanded_acceleration import cyclic_expanded_acceleration 
from .cyclic_expanded_stress import cyclic_expanded_stress 
from .cyclic_expanded_enf import cyclic_expanded_enf 
from .cyclic_volume import cyclic_volume 
from .cyclic_strain_energy import cyclic_strain_energy 
from .migrate_to_h5dpf import migrate_to_h5dpf 



"""
invariants_fc
=============
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.FEMutils plugin, from "invariant" category
"""

class invariants_fc(Operator):
    """Computes the element-wise invariants of all the tensor fields of a fields container.

      available inputs:
        - fields_container (FieldsContainer)

      available outputs:
        - fields_int (FieldsContainer)
        - fields_eqv (FieldsContainer)
        - fields_max_shear (FieldsContainer)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.invariant.invariants_fc()

      >>> # Make input connections
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.invariant.invariants_fc(fields_container=my_fields_container)

      >>> # Get output data
      >>> result_fields_int = op.outputs.fields_int()
      >>> result_fields_eqv = op.outputs.fields_eqv()
      >>> result_fields_max_shear = op.outputs.fields_max_shear()"""
    def __init__(self, fields_container=None, config=None, server=None):
        super().__init__(name="invariants_deriv_fc", config = config, server = server)
        self._inputs = InputsInvariantsFc(self)
        self._outputs = OutputsInvariantsFc(self)
        if fields_container !=None:
            self.inputs.fields_container.connect(fields_container)

    @staticmethod
    def _spec():
        spec = Specification(description="""Computes the element-wise invariants of all the tensor fields of a fields container.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document="""""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "fields_int", type_names=["fields_container"], optional=False, document="""stress intensity field"""), 
                                 1 : PinSpecification(name = "fields_eqv", type_names=["fields_container"], optional=False, document="""stress equivalent intensity"""), 
                                 2 : PinSpecification(name = "fields_max_shear", type_names=["fields_container"], optional=False, document="""max shear stress field""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "invariants_deriv_fc")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsInvariantsFc 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsInvariantsFc 
        """
        return super().outputs


#internal name: invariants_deriv_fc
#scripting name: invariants_fc
class InputsInvariantsFc(_Inputs):
    """Intermediate class used to connect user inputs to invariants_fc operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.invariant.invariants_fc()
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)
    """
    def __init__(self, op: Operator):
        super().__init__(invariants_fc._spec().inputs, op)
        self._fields_container = Input(invariants_fc._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator

        Parameters
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.invariant.invariants_fc()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> #or
        >>> op.inputs.fields_container(my_fields_container)

        """
        return self._fields_container

class OutputsInvariantsFc(_Outputs):
    """Intermediate class used to get outputs from invariants_fc operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.invariant.invariants_fc()
      >>> # Connect inputs : op.inputs. ...
      >>> result_fields_int = op.outputs.fields_int()
      >>> result_fields_eqv = op.outputs.fields_eqv()
      >>> result_fields_max_shear = op.outputs.fields_max_shear()
    """
    def __init__(self, op: Operator):
        super().__init__(invariants_fc._spec().outputs, op)
        self._fields_int = Output(invariants_fc._spec().output_pin(0), 0, op) 
        self._outputs.append(self._fields_int)
        self._fields_eqv = Output(invariants_fc._spec().output_pin(1), 1, op) 
        self._outputs.append(self._fields_eqv)
        self._fields_max_shear = Output(invariants_fc._spec().output_pin(2), 2, op) 
        self._outputs.append(self._fields_max_shear)

    @property
    def fields_int(self):
        """Allows to get fields_int output of the operator


        - pindoc: stress intensity field

        Returns
        ----------
        my_fields_int : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.invariant.invariants_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_int = op.outputs.fields_int() 
        """
        return self._fields_int

    @property
    def fields_eqv(self):
        """Allows to get fields_eqv output of the operator


        - pindoc: stress equivalent intensity

        Returns
        ----------
        my_fields_eqv : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.invariant.invariants_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_eqv = op.outputs.fields_eqv() 
        """
        return self._fields_eqv

    @property
    def fields_max_shear(self):
        """Allows to get fields_max_shear output of the operator


        - pindoc: max shear stress field

        Returns
        ----------
        my_fields_max_shear : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.invariant.invariants_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_max_shear = op.outputs.fields_max_shear() 
        """
        return self._fields_max_shear




import os
import socket

from ansys.dpf.core._version import __version__

# environment variables for pyansys.com
if "jupyter" in socket.gethostname():
    if "ANSYS_PATH" not in os.environ:
        os.environ["ANSYS_PATH"] = "/mnt/ansys_inc/v212/"
    if "DPF_PATH" not in os.environ:
        os.environ["DPF_PATH"] = (
            "/mnt/ansys_inc/dpf/bin_v%s/Ans.dpf.core.Grpc.exe" % __version__
        )
    if "AWP_UNIT_TEST_FILES" not in os.environ:
        os.environ["AWP_UNIT_TEST_FILES"] = "/mnt/ansys_inc/dpf/test_files/"

from ansys.dpf.core.misc import module_exists, Report
from ansys.dpf.core.dpf_operator import Operator, Config
from ansys.dpf.core.model import Model
from ansys.dpf.core.field import Field, FieldDefinition
from ansys.dpf.core.dimensionality import Dimensionality
from ansys.dpf.core.property_field import PropertyField
from ansys.dpf.core.fields_container import FieldsContainer
from ansys.dpf.core.meshes_container import MeshesContainer
from ansys.dpf.core.scopings_container import ScopingsContainer
from ansys.dpf.core.server import (
    start_local_server,
    _global_server,
    connect_to_server,
    has_local_server,
)
from ansys.dpf.core.data_sources import DataSources
from ansys.dpf.core.scoping import Scoping
from ansys.dpf.core.common import types, natures, locations, shell_layers
from ansys.dpf.core import help
from ansys.dpf.core.core import (
    BaseService,
    load_library,
    download_file,
    upload_file,
    upload_file_in_tmp_folder,
    upload_files_in_folder,
    download_files_in_folder,
    make_tmp_dir_server,
)
from ansys.dpf.core.time_freq_support import TimeFreqSupport
from ansys.dpf.core.meshed_region import MeshedRegion
from ansys.dpf.core.elements import element_types
from ansys.dpf.core.result_info import ResultInfo
from ansys.dpf.core.collection import Collection
from ansys.dpf.core.workflow import Workflow
from ansys.dpf.core.cyclic_support import CyclicSupport
from ansys.dpf.core.element_descriptor import ElementDescriptor
from ansys.dpf.core import operators
from ansys.dpf.core.fields_factory import field_from_array
from ansys.dpf.core import (
    fields_container_factory,
    fields_factory,
    mesh_scoping_factory,
    time_freq_scoping_factory,
)
from ansys.dpf.core import server
from ansys.dpf.core import check_version

# for matplotlib
# solves "QApplication: invalid style override passed, ignoring it."
os.environ["QT_STYLE_OVERRIDE"] = ""

# Setup data directory
USER_DATA_PATH = None
EXAMPLES_PATH = None
if os.environ.get("DPF_DOCKER", False):  # pragma: no cover
    # Running DPF within docker (likely for CI)
    # path must be relative to DPF directory
    #
    # assumes the following docker mount:
    # -v /tmp:/dpf/_cache
    EXAMPLES_PATH = "/tmp"
else:
    try:
        import appdirs

        USER_DATA_PATH = appdirs.user_data_dir("ansys-dpf-core")
        if not os.path.exists(USER_DATA_PATH):  # pragma: no cover
            os.makedirs(USER_DATA_PATH)

        EXAMPLES_PATH = os.path.join(USER_DATA_PATH, "examples")
        if not os.path.exists(EXAMPLES_PATH):  # pragma: no cover
            os.makedirs(EXAMPLES_PATH)
    except:  # pragma: no cover
        pass


# Configure PyVista's ``rcParams`` for dpf
if module_exists("pyvista"):
    import pyvista as pv

    pv.rcParams["interactive"] = True
    pv.rcParams["cmap"] = "jet"
    pv.rcParams["font"]["family"] = "courier"
    pv.rcParams["title"] = "DPF"


SERVER = None

_server_instances = []

def disable_off_screen_rendering() -> None:
    # enable matplotlib off_screen plotting to avoid test interruption
    if module_exists("matplotlib"):
        import matplotlib as mpl

        mpl.use("Agg")

    # enable off_screen plotting to avoid test interruption
    if module_exists("pyvista"):
        import pyvista as pv

        pv.OFF_SCREEN = True



"""
accumulate_min_over_label_fc
============================
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.Native plugin, from "math" category
"""

class accumulate_min_over_label_fc(Operator):
    """Compute the component-wise sum over all the fields having the same id for the label set in input in the fields container and take its opposite. This computation can be incremental, if the input fields container is connected and the operator is ran several time, the output field will be on all the inputs connected

      available inputs:
        - fields_container (FieldsContainer)

      available outputs:
        - field (Field)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.math.accumulate_min_over_label_fc()

      >>> # Make input connections
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.math.accumulate_min_over_label_fc(fields_container=my_fields_container)

      >>> # Get output data
      >>> result_field = op.outputs.field()"""
    def __init__(self, fields_container=None, config=None, server=None):
        super().__init__(name="accumulate_min_over_label_fc", config = config, server = server)
        self._inputs = InputsAccumulateMinOverLabelFc(self)
        self._outputs = OutputsAccumulateMinOverLabelFc(self)
        if fields_container !=None:
            self.inputs.fields_container.connect(fields_container)

    @staticmethod
    def _spec():
        spec = Specification(description="""Compute the component-wise sum over all the fields having the same id for the label set in input in the fields container and take its opposite. This computation can be incremental, if the input fields container is connected and the operator is ran several time, the output field will be on all the inputs connected""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document="""""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "field", type_names=["field"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "accumulate_min_over_label_fc")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsAccumulateMinOverLabelFc 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsAccumulateMinOverLabelFc 
        """
        return super().outputs


#internal name: accumulate_min_over_label_fc
#scripting name: accumulate_min_over_label_fc
class InputsAccumulateMinOverLabelFc(_Inputs):
    """Intermediate class used to connect user inputs to accumulate_min_over_label_fc operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.math.accumulate_min_over_label_fc()
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)
    """
    def __init__(self, op: Operator):
        super().__init__(accumulate_min_over_label_fc._spec().inputs, op)
        self._fields_container = Input(accumulate_min_over_label_fc._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator

        Parameters
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.accumulate_min_over_label_fc()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> #or
        >>> op.inputs.fields_container(my_fields_container)

        """
        return self._fields_container

class OutputsAccumulateMinOverLabelFc(_Outputs):
    """Intermediate class used to get outputs from accumulate_min_over_label_fc operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.math.accumulate_min_over_label_fc()
      >>> # Connect inputs : op.inputs. ...
      >>> result_field = op.outputs.field()
    """
    def __init__(self, op: Operator):
        super().__init__(accumulate_min_over_label_fc._spec().outputs, op)
        self._field = Output(accumulate_min_over_label_fc._spec().output_pin(0), 0, op) 
        self._outputs.append(self._field)

    @property
    def field(self):
        """Allows to get field output of the operator


        Returns
        ----------
        my_field : Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.accumulate_min_over_label_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field = op.outputs.field() 
        """
        return self._field




"""
.. _ref_scoping:

Scoping
=======
"""

import array
import sys

import numpy as np
from ansys.dpf.core.check_version import server_meet_version, version_requires
from ansys.dpf.core.common import _common_progress_bar, locations
from ansys.dpf.core.misc import DEFAULT_FILE_CHUNK_SIZE
from ansys.grpc.dpf import base_pb2, scoping_pb2, scoping_pb2_grpc


class Scoping:
    """Represents a scoping, which is a subset of a model support.

    Parameters
    ----------
    scoping : ansys.grpc.dpf.scoping_pb2.Scoping message, optional
    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.
    server : DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    Attributes
    ----------
    ids : list of int
        List of IDs to include in the scoping.
    location : str
        Location of the IDs, such as ``"Nodal"`` or ``"Elemental"``.

    Examples
    --------
    Create a mesh scoping.

    >>> from ansys.dpf import core as dpf
    >>> # 1. using the mesh_scoping_factory
    >>> from ansys.dpf.core import mesh_scoping_factory
    >>> # a. scoping with elemental location that targets the elements with id 2, 7 and 11
    >>> my_elemental_scoping = mesh_scoping_factory.elemental_scoping([2, 7, 11])
    >>> # b. scoping with nodal location that targets the elements with id 4 and 6
    >>> my_nodal_scoping = mesh_scoping_factory.nodal_scoping([4, 6])
    >>> #2. using the classic API
    >>> my_scoping = dpf.Scoping()
    >>> my_scoping.location = "Nodal" #optional
    >>> my_scoping.ids = list(range(1,11))

    """

    def __init__(self, scoping=None, server=None, ids=None, location=None):
        """Initializes the scoping with an optional scoping message or
        by connecting to a stub.
        """
        if server is None:
            import ansys.dpf.core.server as serverlib

            server = serverlib._global_server()

        self._server = server
        self._stub = self._connect()

        if scoping is None:
            request = base_pb2.Empty()
            self._message = self._stub.Create(request)
        else:
            self._message = scoping

        if ids:
            self.ids = ids
        if location:
            self.location = location

    def _count(self):
        """
        Returns
        -------
        count : int
            Number of scoping IDs.
        """
        request = scoping_pb2.CountRequest()
        request.entity = base_pb2.NUM_ELEMENTARY_DATA
        request.scoping.CopyFrom(self._message)
        return self._stub.Count(request).count

    def _get_location(self):
        """Retrieve the location of the IDs.

        Returns
        -------
        location : str
            Location of the IDs.
        """
        sloc = self._stub.GetLocation(self._message).loc.location
        return sloc

    def _set_location(self, loc=locations.nodal):
        """
        Parameters
        ----------
        loc : str or core.locations enum
            Location needed.

        """
        request = scoping_pb2.UpdateRequest()
        request.location.location = loc
        request.scoping.CopyFrom(self._message)
        self._stub.Update(request)

    @version_requires("2.1")
    def _set_ids(self, ids):
        """
        Parameters
        ----------
        ids : list of int
            IDs to set.

        Notes
        -----
        Print a progress bar.
        """
        # must convert to a list for gRPC
        if isinstance(ids, range):
            ids = np.array(list(ids), dtype=np.int32)
        elif not isinstance(ids, (np.ndarray, np.generic)):
            ids = np.array(ids, dtype=np.int32)
        else:
            ids = np.array(list(ids), dtype=np.int32)

        metadata = [("size_int", f"{len(ids)}")]
        request = scoping_pb2.UpdateIdsRequest()
        request.scoping.CopyFrom(self._message)
        if server_meet_version("2.1", self._server):
            self._stub.UpdateIds(_data_chunk_yielder(request, ids), metadata=metadata)
        else:
            self._stub.UpdateIds(
                _data_chunk_yielder(request, ids, 8.0e6), metadata=metadata
            )

    def _get_ids(self, np_array=False):
        """
        Returns
        -------
        ids : list[int], numpy.array (if np_array==True)
            List of IDs.

        Notes
        -----
        Print a progress bar.
        """
        if server_meet_version("2.1", self._server):
            service = self._stub.List(self._message)
            dtype = np.int32
            return _data_get_chunk_(dtype, service, np_array)
        else:
            out = []

            service = self._stub.List(self._message)
            for chunk in service:
                out.extend(chunk.ids.rep_int)
            if np_array:
                return np.array(out, dtype=np.int32)
            else:
                return out

    def set_id(self, index, scopingid):
        """Set the ID of a scoping's index.

        Parameters
        ----------
        index : int
            Index of the scoping.
        scopingid : int
            ID of the scoping.
        """
        request = scoping_pb2.UpdateRequest()
        request.index_id.id = scopingid
        request.index_id.index = index
        request.scoping.CopyFrom(self._message)
        self._stub.Update(request)

    def _get_id(self, index):
        """Retrieve the index that the scoping ID is located on.

        Parameters
        ----------
        index : int
            Index of the scoping

        Returns
        -------
        id : int
            ID of the scoping's index.
        """
        request = scoping_pb2.GetRequest()
        request.index = index
        request.scoping.CopyFrom(self._message)
        return self._stub.Get(request).id

    def _get_index(self, scopingid):
        """Retrieve an ID corresponding to an ID in the scoping.

        Parameters
        ----------
        id : int
            ID to retrieve.

        Returns
        -------
        index : int
            Index of the ID.
        """
        request = scoping_pb2.GetRequest()
        request.id = scopingid
        request.scoping.CopyFrom(self._message)
        return self._stub.Get(request).index

    def id(self, index: int):
        """Retrieve the ID at a given index.

        Parameters
        ----------
        index : int
            Index for the ID.

        Returns
        -------
        size : int

        """
        return self._get_id(index)

    def index(self, id: int):
        """Retrieve the index of a given ID.

        Parameters
        ----------
        id : int
            ID for the index to retrieve.

        Returns
        -------
        size : int

        """
        return self._get_index(id)

    @property
    def ids(self):
        """Retrive a list of IDs in the scoping.

        Returns
        -------
        ids : list of int
            List of IDs to retrieve.

        Notes
        -----
        Print a progress bar.
        """
        return self._get_ids()

    @ids.setter
    def ids(self, value):
        self._set_ids(value)

    @property
    def location(self):
        """Location of the IDs as a string, such as ``"nodal"``, ``"elemental"``,
        and ``"time_freq"``.

        Returns
        -------
        location : str

        """
        return self._get_location()

    @location.setter
    def location(self, value):
        self._set_location(value)

    def _connect(self):
        """Connect to the gRPC service containing the reader."""
        return scoping_pb2_grpc.ScopingServiceStub(self._server.channel)

    def __len__(self):
        return self._count()

    def __del__(self):
        try:
            self._stub.Delete(self._message)
        except:
            pass

    def __iter__(self):
        return self.ids.__iter__()

    def __getitem__(self, key):
        """Retrieve the ID at a requested index."""
        return self.id(key)

    @property
    def size(self):
        """Length of the list of IDs.

        Returns
        -------
        size : int

        """
        return self._count()

    def __str__(self):
        """Describe the entity.

        Returns
        -------
        description : str
        """
        from ansys.dpf.core.core import _description

        return _description(self._message, self._server)

    def deep_copy(self, server=None):
        """Create a deep copy of the scoping's data on a given server.

        This method is useful for passiong data from one server instance to another.

        Parameters
        ----------
        server : ansys.dpf.core.server, optional
            Server with the channel connected to the remote or local instance.
            The default is ``None``, in which case an attempt is made to use the
            global server.

        Returns
        -------
        scoping_copy : Scoping
        """
        scop = Scoping(server=server)
        scop.ids = self.ids
        scop.location = self.location
        return scop


def _data_chunk_yielder(request, data, chunk_size=DEFAULT_FILE_CHUNK_SIZE):
    length = data.size
    need_progress_bar = length > 1e6
    if need_progress_bar:
        bar = _common_progress_bar(
            "Sending data...", unit=data.dtype.name, tot_size=length
        )
        bar.start()
    sent_length = 0
    if length == 0:
        yield request
        return
    unitary_size = int(chunk_size // sys.getsizeof(data[0]))
    if length - sent_length < unitary_size:
        unitary_size = length - sent_length
    while sent_length < length:
        currentcopy = data.take(range(sent_length, sent_length + unitary_size))
        request.array = currentcopy.tobytes()
        sent_length = sent_length + unitary_size
        if length - sent_length < unitary_size:
            unitary_size = length - sent_length
        yield request
        try:
            if need_progress_bar:
                bar.update(sent_length)
        except:
            pass
    try:
        if need_progress_bar:
            bar.finish()
    except:
        pass


def _data_get_chunk_(dtype, service, np_array=True):
    tupleMetaData = service.initial_metadata()

    need_progress_bar = False
    for iMeta in range(len(tupleMetaData)):
        if tupleMetaData[iMeta].key == "size_tot":
            size = int(tupleMetaData[iMeta].value)

    itemsize = np.dtype(dtype).itemsize
    need_progress_bar = size // itemsize > 1e6
    if need_progress_bar:
        bar = _common_progress_bar(
            "Receiving data...", unit=dtype.__name__ + "s", tot_size=size // itemsize
        )
        bar.start()

    if np_array:
        arr = np.empty(size // itemsize, dtype)
        i = 0
        for chunk in service:
            curr_size = len(chunk.array) // itemsize
            arr[i : i + curr_size] = np.frombuffer(chunk.array, dtype)
            i += curr_size
            try:
                if need_progress_bar:
                    bar.update(i)
            except:
                pass

    else:
        arr = []
        if dtype == np.float:
            dtype = "d"
        else:
            dtype = "i"
        for chunk in service:
            arr.extend(array.array(dtype, chunk.array))
            try:
                if need_progress_bar:
                    bar.update(len(arr))
            except:
                pass
    try:
        if need_progress_bar:
            bar.finish()
    except:
        pass
    return arr



"""
to_polar_coordinates
====================
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.FEMutils plugin, from "geo" category
"""

class to_polar_coordinates(Operator):
    """Find r, theta (rad), z coordinates of a coordinates (nodal) field in cartesian coordinates system with respoect to the input coordinate system defining the rotation axis and the origin.

      available inputs:
        - field (Field, FieldsContainer)
        - coordinate_system (Field) (optional)

      available outputs:
        - fields_container (FieldsContainer)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.geo.to_polar_coordinates()

      >>> # Make input connections
      >>> my_field = dpf.Field()
      >>> op.inputs.field.connect(my_field)
      >>> my_coordinate_system = dpf.Field()
      >>> op.inputs.coordinate_system.connect(my_coordinate_system)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.geo.to_polar_coordinates(field=my_field,coordinate_system=my_coordinate_system)

      >>> # Get output data
      >>> result_fields_container = op.outputs.fields_container()"""
    def __init__(self, field=None, coordinate_system=None, config=None, server=None):
        super().__init__(name="polar_coordinates", config = config, server = server)
        self._inputs = InputsToPolarCoordinates(self)
        self._outputs = OutputsToPolarCoordinates(self)
        if field !=None:
            self.inputs.field.connect(field)
        if coordinate_system !=None:
            self.inputs.coordinate_system.connect(coordinate_system)

    @staticmethod
    def _spec():
        spec = Specification(description="""Find r, theta (rad), z coordinates of a coordinates (nodal) field in cartesian coordinates system with respoect to the input coordinate system defining the rotation axis and the origin.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "field", type_names=["field","fields_container"], optional=False, document="""field or fields container with only one field is expected"""), 
                                 1 : PinSpecification(name = "coordinate_system", type_names=["field"], optional=True, document="""3-3 rotation matrix and origin coordinates must be set here to define a coordinate system. By default, the rotation axis is the z axis and the origin is [0,0,0]""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "polar_coordinates")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsToPolarCoordinates 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsToPolarCoordinates 
        """
        return super().outputs


#internal name: polar_coordinates
#scripting name: to_polar_coordinates
class InputsToPolarCoordinates(_Inputs):
    """Intermediate class used to connect user inputs to to_polar_coordinates operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.geo.to_polar_coordinates()
      >>> my_field = dpf.Field()
      >>> op.inputs.field.connect(my_field)
      >>> my_coordinate_system = dpf.Field()
      >>> op.inputs.coordinate_system.connect(my_coordinate_system)
    """
    def __init__(self, op: Operator):
        super().__init__(to_polar_coordinates._spec().inputs, op)
        self._field = Input(to_polar_coordinates._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._field)
        self._coordinate_system = Input(to_polar_coordinates._spec().input_pin(1), 1, op, -1) 
        self._inputs.append(self._coordinate_system)

    @property
    def field(self):
        """Allows to connect field input to the operator

        - pindoc: field or fields container with only one field is expected

        Parameters
        ----------
        my_field : Field, FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.geo.to_polar_coordinates()
        >>> op.inputs.field.connect(my_field)
        >>> #or
        >>> op.inputs.field(my_field)

        """
        return self._field

    @property
    def coordinate_system(self):
        """Allows to connect coordinate_system input to the operator

        - pindoc: 3-3 rotation matrix and origin coordinates must be set here to define a coordinate system. By default, the rotation axis is the z axis and the origin is [0,0,0]

        Parameters
        ----------
        my_coordinate_system : Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.geo.to_polar_coordinates()
        >>> op.inputs.coordinate_system.connect(my_coordinate_system)
        >>> #or
        >>> op.inputs.coordinate_system(my_coordinate_system)

        """
        return self._coordinate_system

class OutputsToPolarCoordinates(_Outputs):
    """Intermediate class used to get outputs from to_polar_coordinates operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.geo.to_polar_coordinates()
      >>> # Connect inputs : op.inputs. ...
      >>> result_fields_container = op.outputs.fields_container()
    """
    def __init__(self, op: Operator):
        super().__init__(to_polar_coordinates._spec().outputs, op)
        self._fields_container = Output(to_polar_coordinates._spec().output_pin(0), 0, op) 
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator


        Returns
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.geo.to_polar_coordinates()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container() 
        """
        return self._fields_container




from .serializer import serializer 
from .mechanical_csv_to_field import mechanical_csv_to_field 
from .field_to_csv import field_to_csv 
from .deserializer import deserializer 
from .csv_to_field import csv_to_field 
from .vtk_export import vtk_export 
from .vtk_to_fields import vtk_to_fields 
from .migrate_file_to_vtk import migrate_file_to_vtk 
from .serialize_to_hdf5 import serialize_to_hdf5 



"""
Collection
===========
Contains classes associated with the DPF collection.

"""

from ansys import dpf
from ansys.grpc.dpf import collection_pb2, collection_pb2_grpc
from ansys.dpf.core.core import base_pb2
from ansys.dpf.core.common import types
from ansys.dpf.core.scoping import Scoping, scoping_pb2
from ansys.dpf.core.field import Field, field_pb2
from ansys.dpf.core.meshed_region import MeshedRegion, meshed_region_pb2
from ansys.dpf.core.time_freq_support import TimeFreqSupport
from ansys.dpf.core.errors import protect_grpc
from ansys.dpf.core import server
from ansys.dpf.core.scoping import Scoping


class Collection:
    """Represents a collection of entries ordered by labels and IDs.

    Parameters
    ----------
    dpf_type :

    collection : ansys.grpc.dpf.collection_pb2.Collection, optional
        Collection to create from the collection message. The default is ``None``.
    server : server.DPFServer, optional
        Server with the channel connected to the remote or local instance. The
        default is ``None``, in which case an attempt is made to use the global
        server.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> coll = dpf.Collection(dpf.types.field)

    """

    def __init__(self, dpf_type, collection=None, server: server.DpfServer = None):
        if server is None:
            server = dpf.core._global_server()

        self._server = server
        self._stub = self._connect()
        self._type = dpf_type
        # self.__info = None  # cached info

        if collection is None:
            request = collection_pb2.CollectionRequest()
            if hasattr(dpf_type, "name"):
                stype = dpf_type.name
            else:
                stype = dpf_type
            request.type = base_pb2.Type.Value(stype.upper())
            self._message = self._stub.Create(request)
        elif hasattr(collection, "_message"):
            self._message = collection._message
            self._collection = collection  # keep the base collection used for copy
        else:
            self._message = collection

    def set_labels(self, labels):
        """Set labels for scoping the collection.

        Parameters
        ----------
        labels : list[str], optional
            Labels to scope entries to. For example, ``["time", "complex"]``.

        """
        if len(self._info["labels"]) != 0:
            print(
                "The collection already has labels :",
                self._info["labels"],
                "deleting existing labels is not implemented yet.",
            )
            return
        request = collection_pb2.UpdateLabelsRequest()
        request.collection.CopyFrom(self._message)
        request.labels.extend([collection_pb2.NewLabel(label=lab) for lab in labels])
        self._stub.UpdateLabels(request)

    def add_label(self, label, default_value=None):
        """Add the requested label to scope the collection.

        Parameters
        ----------
        label : str
            Labels to scope the etnries to. For example, ``"time"``.
        default_value : int, optional
            Default value for existing fields in the collection. The default
            is ``None``.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> coll = dpf.Collection(dpf.types.field)
        >>> coll.add_label('time')

        """
        request = collection_pb2.UpdateLabelsRequest()
        request.collection.CopyFrom(self._message)
        new_label = collection_pb2.NewLabel(label=label)
        if default_value is not None:
            new_label.default_value.default_value = default_value
        request.labels.extend([new_label])
        self._stub.UpdateLabels(request)

    def _get_labels(self):
        """Retrieve labels scoping the collection.

        Returns
        -------
        labels: list[str]
            List of labels that entries are scoped to. For example, ``["time", "complex"]``.
        """
        return self._info["labels"]

    labels = property(_get_labels, set_labels, "labels")

    def has_label(self, label):
        """Check if a collection has a specified label.

        Parameters
        ----------
        label: str
            Label to search for. For example, ``"time"``.

        Returns
        -------
        bool
           ``True`` when successful, ``False`` when failed.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> coll = dpf.Collection(dpf.types.field)
        >>> coll.add_label('time')
        >>> coll.has_label('time')
        True

        >>> coll.has_label('complex')
        False

        """
        return label in self.labels

    def _get_entries(self, label_space_or_index):
        """Retrieve the entries at a requested label space or index.

        Parameters
        ----------
        label_space_or_index : dict[str,int]
            Label space or index. For example,
            ``{"time": 1, "complex": 0}`` or the index of the field.

        Returns
        -------
        entries : list[Scoping], list[Field], list[MeshedRegion]
            Entries corresponding to the request.
        """
        request = collection_pb2.EntryRequest()
        request.collection.CopyFrom(self._message)

        if isinstance(label_space_or_index, dict):
            for key in label_space_or_index:
                request.label_space.label_space[key] = label_space_or_index[key]
        elif isinstance(label_space_or_index, int):
            request.index = label_space_or_index

        out = self._stub.GetEntries(request)
        list_out = []
        for obj in out.entries:
            if obj.HasField("dpf_type"):
                if self._type == types.scoping:
                    unpacked_msg = scoping_pb2.Scoping()
                    obj.dpf_type.Unpack(unpacked_msg)
                    list_out.append(Scoping(scoping=unpacked_msg, server=self._server))
                elif self._type == types.field:
                    unpacked_msg = field_pb2.Field()
                    obj.dpf_type.Unpack(unpacked_msg)
                    list_out.append(Field(field=unpacked_msg, server=self._server))
                elif self._type == types.meshed_region:
                    unpacked_msg = meshed_region_pb2.MeshedRegion()
                    obj.dpf_type.Unpack(unpacked_msg)
                    list_out.append(
                        MeshedRegion(mesh=unpacked_msg, server=self._server)
                    )
        if len(list_out) == 0:
            list_out = None
        return list_out

    def _get_entry(self, label_space_or_index):
        """Retrieve the entry at a requested label space or index.

        Parameters
        ----------
        label_space_or_index : dict[str,int]
            Label space or index of the requested entry. For example,
            ``{"time": 1, "complex": 0}`` or the index of the field.

        Returns
        -------
        entry : Scoping, Field, MeshedRegion
            Entry at the requested label space or index.
        """
        entries = self._get_entries(label_space_or_index)
        if isinstance(entries, list):
            if len(entries) == 1:
                return entries[0]
            else:
                raise KeyError(f"{label_space_or_index} has {len(entries)} entries")
        else:
            return entries

    def get_label_space(self, index):
        """Retrieve the label space of an entry at a requested index.

        Parameters
        ----------
        index: int
            Index of the entry.

        Returns
        -------
        label_space : dict(str:int)
            Scoping of the requested entry. For example,
            ``{"time": 1, "complex": 0}``.
        """
        request = collection_pb2.EntryRequest()
        request.collection.CopyFrom(self._message)
        request.index = index
        out = self._stub.GetEntries(request).entries
        out = out[0].label_space.label_space
        dictOut = {}
        for key in out:
            dictOut[key] = out[key]

        return dictOut

    def get_available_ids_for_label(self, label="time"):
        """Retrieve the IDs assigned to an input label.


        Parameters
        ----------
        label : str
            Name of the input label. The default is ``"time"``.

        Returns
        -------
        ids : list[int]
            List of IDs assigned to the input label.
        """
        ids = []
        for i in range(len(self)):
            current_scop = self.get_label_space(i)
            if label in current_scop and current_scop[label] not in ids:
                ids.append(current_scop[label])
        return ids

    def get_label_scoping(self, label="time"):
        """Retrieve the scoping for an input label.

        This method allows you to retrieve a list of IDs for a given input label in the
        collection. For example, if the label ``el_type`` exists in the collection, you
        can use the `get_lable_scoping` method to retrieve a list of IDS with this label.
        You can then use these IDs to request a given entity inside the collection.

        Parameters
        ----------
        label: str
            Name of the input label.

        Returns
        -------
        scoping: Scoping
            IDs scopped to the input label.
        """
        request = collection_pb2.LabelScopingRequest()
        request.collection.CopyFrom(self._message)
        request.label = label
        scoping_message = self._stub.GetLabelScoping(request)
        scoping = Scoping(scoping_message.label_scoping)
        return scoping

    def __getitem__(self, index):
        """Retrieves the entry at a requested index value.

        Parameters
        ----------
        index : int
            Index value.

        Returns
        -------
        entry : Field , Scoping
            Entry at the index value.
        """
        self_len = len(self)
        if index < 0:
            # convert to a positive index
            index = self_len + index

        if not self_len:
            raise IndexError("This collection contains no items")
        if index >= self_len:
            raise IndexError(f"This collection contains only {self_len} entrie(s)")

        return self._get_entries(index)[0]

    def _add_entry(self, label_space, entry):
        """Update or add an entry at a requested label space.

        parameters
        ----------
        label_space : list[str,int]
            Label space of the requested fields. For example, ``{"time":1, "complex":0}``.
        entry : Field or Scoping
            DPF entry to add.
        """
        request = collection_pb2.UpdateRequest()
        request.collection.CopyFrom(self._message)
        request.entry.dpf_type.Pack(entry._message)
        for key in label_space:
            request.label_space.label_space[key] = label_space[key]
        self._stub.UpdateEntry(request)

    def _get_time_freq_support(self):
        """Retrieve time frequency support.

        Returns
        -------
        time_freq_support : TimeFreqSupport
        """
        request = collection_pb2.SupportRequest()
        request.collection.CopyFrom(self._message)
        request.type = base_pb2.Type.Value("TIME_FREQ_SUPPORT")
        message = self._stub.GetSupport(request)
        return TimeFreqSupport(time_freq_support=message)

    def _set_time_freq_support(self, time_freq_support):
        """Set the time frequency support of the collection."""
        request = collection_pb2.UpdateSupportRequest()
        request.collection.CopyFrom(self._message)
        request.time_freq_support.CopyFrom(time_freq_support._message)
        request.label = "time"
        self._stub.UpdateSupport(request)

    def _connect(self):
        """Connect to the gRPC service."""
        return collection_pb2_grpc.CollectionServiceStub(self._server.channel)

    @property
    @protect_grpc
    def _info(self):
        """Length and labels of this container."""
        list_stub = self._stub.List(self._message)
        return {"len": list_stub.count_entries, "labels": list_stub.labels.labels}

    def __str__(self):
        """Describe the entity.

        Returns
        -------
        description : str
            Description of the entity.
        """
        request = base_pb2.DescribeRequest()
        request.dpf_type_id = self._message.id
        return self._stub.Describe(request).description

    def __len__(self):
        """Retrieve the number of entries."""
        return self._info["len"]

    def __del__(self):
        """Delete the entry."""
        try:
            self._stub.Delete(self._message)
        except:
            pass

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]



from .identical_meshes import identical_meshes 
from .component_selector_fc import component_selector_fc 
from .component_selector import component_selector 
from .identical_property_fields import identical_property_fields 
from .merge_fields_by_label import merge_fields_by_label 
from .solid_shell_fields import solid_shell_fields 
from .identical_fields import identical_fields 
from .included_fields import included_fields 
from .identical_fc import identical_fc 
from .enrich_materials import enrich_materials 



from .eigen_values import eigen_values 
from .principal_invariants import principal_invariants 
from .von_mises_eqv import von_mises_eqv 
from .von_mises_eqv_fc import von_mises_eqv_fc 
from .invariants import invariants 
from .eigen_values_fc import eigen_values_fc 
from .invariants_fc import invariants_fc 
from .principal_invariants_fc import principal_invariants_fc 
from .eigen_vectors_fc import eigen_vectors_fc 
from .eigen_vectors import eigen_vectors 



"""
fields_factory
==============

Contains functions to simplify creating fields.
"""

from ansys.dpf import core
from ansys.dpf.core.common import natures, locations
from ansys.dpf.core import Field
from ansys.dpf.core.dimensionality import Dimensionality
from ansys.grpc.dpf import field_pb2, field_pb2_grpc, base_pb2

import numpy as np


def field_from_array(arr):
    """Create a DPF vector or scalar field from a numpy array or a Python list.

    Parameters
    ----------
    arr : np.ndarray or List
        Numpy array or Python list containing either 1 or 3 dimensions.

    Returns
    -------
    field : Field
        Field constructed from the array.
    """
    from ansys.dpf.core import Field, natures

    arr = np.asarray(arr)

    if not np.issubdtype(arr.dtype, np.number):
        raise TypeError("Array must be a numeric type")

    shp_err = ValueError(
        "Array must be either contain 1 dimension or "
        "2 dimensions with three components."
    )
    if arr.ndim == 1:
        nature = natures.scalar
    elif arr.ndim == 2:
        if arr.shape[1] == 1:
            arr = arr.ravel()
            nature = natures.scalar
        elif arr.shape[1] == 3:
            nature = natures.vector
        elif arr.shape[1] == 6:
            nature = natures.symmatrix
        else:
            raise shp_err
    else:
        raise shp_err

    n_entities = arr.shape[0]
    field = Field(nentities=n_entities, nature=nature)
    field.data = arr
    field.scoping.ids = np.arange(1, n_entities + 1)
    return field


def create_matrix_field(
    num_entities, num_lines, num_col, location=locations.nodal, server=None
):
    """Create a matrix :class:`ansys.dpf.core.Field`.

    This field contain entities that have a matrix format. This is a "reserve" mechanism,
    not a resize one. This means that you need to append data to grow the size of your field.

    Parameters
    ----------
    num_entities : int
        Number of entities to reserve.
    num_lines : int
        Number of matrix line.
    num_col : int
        Number of matrix columns.
    location : str, optional
        Location of the field. The default is ``"Nodal"``. For example:

        - ``"Nodal"``
        - ``"Elemental"``
        - :class:`ansys.dpf.core.natures.elemental_nodal`

    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    field : Field
        DPF field of the requested format.

    Examples
    --------
    Create a field containing 3 matrix entities of a col*lines = 2*5 size with
    a nodal location (default).

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_matrix_field(3, 5, 2)

    """
    return _create_field(
        server, natures.matrix, num_entities, location, num_col, num_lines
    )


def create_3d_vector_field(num_entities, location=locations.nodal, server=None):
    """Create a specific :class:`ansys.dpf.core.Field` with entities that have 3D vector format.

    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow the size of your field.

    Parameters
    ----------
    num_entities : int
        Number of entities to reserve
    location : str, optional
        Location of the field. The default is ``"Nodal"``. For example:

        - ``"Nodal"``
        - ``"Elemental"``
        - ansys.dpf.core.natures.elemental_nodal

    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    field : Field
        DPF field of the requested format.

    Examples
    --------
    Create a field containing 4 3D vector entities with a nodal location (default).

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_3d_vector_field(4)

    """
    return _create_field(server, natures.vector, num_entities, location)


def create_tensor_field(num_entities, location=locations.nodal, server=None):
    """Create a specific :class:`ansys.dpf.core.Field` with entities that have a 3*3 format.

    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow the size of your field.

    Parameters
    ----------
    num_entities : int
        Number of entities to reserve.
    location : str, optional
        Location of the field. The default is ``"Nodal"``. For example:

        - ``"Nodal"``
        - ``"Elemental"``
        - ansys.dpf.core.natures.elemental_nodal

    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    field : Field
        DPF field in the requested format.

    Examples
    --------
    Create a field containing 4 tensor entities with a nodal location (default).

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_tensor_field(4)

    """
    return _create_field(server, natures.symmatrix, num_entities, location)


def create_scalar_field(num_entities, location=locations.nodal, server=None):
    """Create a specific `:class:`ansys.dpf.core.Field` with entities that are scalar.

    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow the size of your field.

    Parameters
    ----------
    num_entities : int
        Number of entities to reserve
    location : str, optional
        Location of the field. The default is ``"Nodal"``. For example:

        - ``"Nodal"``
        - ``"Elemental"``
        - ansys.dpf.core.natures.elemental_nodal

    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    field : Field
        DPF field in the requested format.

    Examples
    --------
    Create a field containing 4 scalars with a nodal location (default).

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_scalar_field(4)

    """
    return _create_field(server, natures.scalar, num_entities, location)


def create_vector_field(num_entities, num_comp, location=locations.nodal, server=None):
    """Create a specific `:class:`ansys.dpf.core.Field` with entities that have a vector format.

    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow the size of your field.

    Parameters
    ----------
    num_entities : int
        Number of entities to reserve.
    num_comp : int
        Number of vector components.
    location : str, optional
        Location of the field. The default is ``"Nodal"``. For example:

        - ``"Nodal"``
        - ``"Elemental"``
        - ansys.dpf.core.natures.elemental_nodal

    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Returns
    -------
    field : Field
        DPF field in the requested format.

    Examples
    --------
    Create a field containing 3 vector entities of 5 components each with a
    nodal location (default).

    >>> from ansys.dpf.core import fields_factory
    >>> field = fields_factory.create_vector_field(3, 5)

    """
    return _create_field(
        server, natures.vector, num_entities, location, ncomp_n=num_comp
    )


def _connect(server):
    """Connect to the gRPC instance."""
    if server is None:
        server = core._global_server()
    stub = field_pb2_grpc.FieldServiceStub(server.channel)
    return stub


def _create_field(
    server, nature, nentities, location=locations.nodal, ncomp_n=0, ncomp_m=0
):
    """Create a specific :class:`ansys.dpf.core.Field`.

    This is a "reserve" mechanism, not a resize one. This means that you
    need to append data to grow the size of your field.

    Parameters
    ----------
    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.
    snature : str
        Nature of the field entity data. For example:

        - :class:`ansys.dpf.core.natures.matrix`
        - :class:`ansys.dpf.core.natures.scalar`

    num_entities : int
        Number of entities to reserve.
    location : str optional
        Location of the field.  For example:

        - ``"Nodal"``
        - ``"Elemental"``
        - ansys.dpf.core.locations.elemental_nodal

    ncomp_n : int
        Number of lines.
    ncomp_m : int
        Number of columns.

    Returns
    -------
    field : Field
        DPF field in the requested format.
    """
    # ncomp_n is number of column components
    # ncomp_m is number of line components
    # connect to grpc
    stub = _connect(server)
    # set nature
    if hasattr(nature, "name"):
        snature = nature.name
    else:
        snature = nature

    if snature == natures.vector.name:
        elem_data_size = 3
    elif snature == natures.symmatrix.name:
        elem_data_size = 6
    elif snature == natures.scalar.name:
        elem_data_size = 1
    elif snature == natures.matrix.name:
        elem_data_size = ncomp_n * ncomp_m
    else:
        elem_data_size = ncomp_n
    if ncomp_n != 0 and ncomp_m != 0:
        dimensionality = Dimensionality([ncomp_n, ncomp_m], nature)
    elif ncomp_n != 0 and ncomp_m == 0:
        dimensionality = Dimensionality([ncomp_n], nature)
    else:
        dimensionality = None
    # set request
    request = field_pb2.FieldRequest()
    nature = base_pb2.Nature.Value(snature.upper())
    request.nature = nature
    request.location.location = location
    request.size.scoping_size = nentities
    if dimensionality is not None:
        request.dimensionality.CopyFrom(dimensionality._parse_dim_to_message())
    request.size.data_size = nentities * elem_data_size
    # get field
    message = stub.Create(request)
    field = Field(field=message, server=server)
    return field



from .node_coordinates import node_coordinates 
from .from_field import from_field 
from .mesh_provider import mesh_provider 
from .points_from_coordinates import points_from_coordinates 
from .split_mesh import split_mesh 
from .from_scoping import from_scoping 
from .split_fields import split_fields 
from .tri_mesh_skin import tri_mesh_skin 
from .mesh_cut import mesh_cut 
from .external_layer import external_layer 
from .skin import skin 
from .stl_export import stl_export 



"""
overall_dot
===========
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.Native plugin, from "math" category
"""

class overall_dot(Operator):
    """Compute a sdot product between two fields and return a scalar.

      available inputs:
        - FieldA (Field)
        - FieldB (Field)

      available outputs:
        - field (Field)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.math.overall_dot()

      >>> # Make input connections
      >>> my_FieldA = dpf.Field()
      >>> op.inputs.FieldA.connect(my_FieldA)
      >>> my_FieldB = dpf.Field()
      >>> op.inputs.FieldB.connect(my_FieldB)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.math.overall_dot(FieldA=my_FieldA,FieldB=my_FieldB)

      >>> # Get output data
      >>> result_field = op.outputs.field()"""
    def __init__(self, FieldA=None, FieldB=None, config=None, server=None):
        super().__init__(name="native::overall_dot", config = config, server = server)
        self._inputs = InputsOverallDot(self)
        self._outputs = OutputsOverallDot(self)
        if FieldA !=None:
            self.inputs.FieldA.connect(FieldA)
        if FieldB !=None:
            self.inputs.FieldB.connect(FieldB)

    @staticmethod
    def _spec():
        spec = Specification(description="""Compute a sdot product between two fields and return a scalar.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "FieldA", type_names=["field"], optional=False, document=""""""), 
                                 1 : PinSpecification(name = "FieldB", type_names=["field"], optional=False, document="""""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "field", type_names=["field"], optional=False, document="""Field defined on over-all location, contains a unique scalar value""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "native::overall_dot")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsOverallDot 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsOverallDot 
        """
        return super().outputs


#internal name: native::overall_dot
#scripting name: overall_dot
class InputsOverallDot(_Inputs):
    """Intermediate class used to connect user inputs to overall_dot operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.math.overall_dot()
      >>> my_FieldA = dpf.Field()
      >>> op.inputs.FieldA.connect(my_FieldA)
      >>> my_FieldB = dpf.Field()
      >>> op.inputs.FieldB.connect(my_FieldB)
    """
    def __init__(self, op: Operator):
        super().__init__(overall_dot._spec().inputs, op)
        self._FieldA = Input(overall_dot._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._FieldA)
        self._FieldB = Input(overall_dot._spec().input_pin(1), 1, op, -1) 
        self._inputs.append(self._FieldB)

    @property
    def FieldA(self):
        """Allows to connect FieldA input to the operator

        Parameters
        ----------
        my_FieldA : Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.overall_dot()
        >>> op.inputs.FieldA.connect(my_FieldA)
        >>> #or
        >>> op.inputs.FieldA(my_FieldA)

        """
        return self._FieldA

    @property
    def FieldB(self):
        """Allows to connect FieldB input to the operator

        Parameters
        ----------
        my_FieldB : Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.overall_dot()
        >>> op.inputs.FieldB.connect(my_FieldB)
        >>> #or
        >>> op.inputs.FieldB(my_FieldB)

        """
        return self._FieldB

class OutputsOverallDot(_Outputs):
    """Intermediate class used to get outputs from overall_dot operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.math.overall_dot()
      >>> # Connect inputs : op.inputs. ...
      >>> result_field = op.outputs.field()
    """
    def __init__(self, op: Operator):
        super().__init__(overall_dot._spec().outputs, op)
        self._field = Output(overall_dot._spec().output_pin(0), 0, op) 
        self._outputs.append(self._field)

    @property
    def field(self):
        """Allows to get field output of the operator


        - pindoc: Field defined on over-all location, contains a unique scalar value

        Returns
        ----------
        my_field : Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.math.overall_dot()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field = op.outputs.field() 
        """
        return self._field




"""
Core
====
"""
import os
import logging
import time
import weakref
import pathlib

import grpc

from ansys.grpc.dpf import base_pb2, base_pb2_grpc
from ansys.dpf.core.errors import protect_grpc
from ansys.dpf.core import server as serverlib
from ansys.dpf.core.misc import DEFAULT_FILE_CHUNK_SIZE
from ansys.dpf.core.common import _common_progress_bar

LOG = logging.getLogger(__name__)
LOG.setLevel("DEBUG")

if "DPF_CONFIGURATION" in os.environ:
    CONFIGURATION = os.environ["DPF_CONFIGURATION"]
else:
    CONFIGURATION = "release"


def load_library(filename, name="", symbol="LoadOperators", server=None):
    """Dynamically load an operators library for dpf.core.
    Code containing this library's operators is generated in
    ansys.dpf.core.operators

    Parameters
    ----------
    filename : str
        Filename of the operator library.

    name : str, optional
        Library name.  Probably optional

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    Examples
    --------
    Load the mesh operators for Windows (for Linux, just use
    'libmeshOperatorsCore.so' instead of 'meshOperatorsCore.dll')

    >>> from ansys.dpf import core as dpf
    >>> # dpf.load_library('meshOperatorsCore.dll', 'mesh_operators')

    """
    base = BaseService(server, load_operators=False)
    base.load_library(filename, name, symbol)
    return name + " successfully loaded"


def upload_file_in_tmp_folder(file_path, new_file_name=None, server=None):
    """Upload a file from the client to the server in a temporary folder
    deleted when the server is shutdown

    Parameters
    ----------
    file_path : str
        file path on the client side to upload

    new_file_name : str, optional
        name to give to the file server side,
        if no name is specified, the same name as the input file is given

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    Notes
    -----
    Print a progress bar

    Returns
    -------
      server_file_path : str
           path generated server side

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> file_path = dpf.upload_file_in_tmp_folder(examples.static_rst)

    """
    base = BaseService(server, load_operators=False)
    return base.upload_file_in_tmp_folder(file_path, new_file_name)


def upload_files_in_folder(
    to_server_folder_path, client_folder_path, specific_extension=None, server=None
):
    """Upload all the files from a folder of the client
    to the target server folder path.

    Parameters
    ----------
    to_server_folder_path : str
        folder path target where will be uploaded files on the server side

    client_folder_path: str
        folder path where the files that must be uploaded are located
        on client side

    specific_extension (optional) : str
        copies only the files with the given extension

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    Returns
    -------
    paths : list of str
        new file paths server side
    """
    base = BaseService(server, load_operators=False)
    return base.upload_files_in_folder(
        to_server_folder_path, client_folder_path, specific_extension
    )


def download_file(server_file_path, to_client_file_path, server=None):
    """Download a file from the server to the target client file path

    Parameters
    ----------
    server_file_path : str
        file path to download on the server side

    to_client_file_path: str
        file path target where the file will be located client side

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    Notes
    -----
    Print a progress bar

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> import os
    >>> file_path = dpf.upload_file_in_tmp_folder(examples.static_rst)
    >>> dpf.download_file(file_path, examples.static_rst)

    """
    base = BaseService(server, load_operators=False)
    return base.download_file(server_file_path, to_client_file_path)


def download_files_in_folder(
    server_folder_path, to_client_folder_path, specific_extension=None, server=None
):
    """Download all the files from a folder of the server
    to the target client folder path

    Parameters
    ----------
    server_folder_path : str
        folder path to download on the server side

    to_client_folder_path: str
        folder path target where the files will be located client side

    specific_extension (optional) : str
        copies only the files with the given extension

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    Notes
    -----
    Print a progress bar

    Returns
    -------
    paths : list of str
        new file paths client side
    """
    base = BaseService(server, load_operators=False)
    return base.download_files_in_folder(
        server_folder_path, to_client_folder_path, specific_extension
    )


def upload_file(file_path, to_server_file_path, server=None):
    """Upload a file from the client to the target server file path

    Parameters
    ----------
    file_path : str
        file path on the client side to upload

    to_server_file_path: str
        file path target where the file will be located server side

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    Notes
    -----
    Print a progress bar

    Returns
    -------
    server_file_path : str
        path generated server side
    """
    base = BaseService(server, load_operators=False)
    return base.upload_file(file_path, to_server_file_path)


def make_tmp_dir_server(server=None):
    """Create a temporary folder server side. Only one temporary folder can be created
    by server instance.
    The folder will be deleted when the server is stopped.

    Parameters
    ----------
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    Returns
    -------
    path : str
        path to the temporary dir
    """
    base = BaseService(server, load_operators=False)
    return base.make_tmp_dir_server()


def _description(dpf_entity_message, server=None):
    """Ask the server to describe the entity in input

    Parameters
    ----------
    dpf_entity_message : core.Operator._message, core.Workflow._message,
                         core.Scoping._message, core.Field._message,
                         core.FieldContainer._message, core.MeshedRegion._message

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    Returns
    -------
       description : str
    """
    return BaseService(server, load_operators=False)._description(dpf_entity_message)


class BaseService:
    """The Base Service class allows to make generic requests to dpf's server.
    For example, information about the server can be requested,
    uploading/downloading file from and to the server can be done,
    new operators plugins can be loaded...
    Most of the request done by the BaseService class are wrapped by
    functions.

    Parameters
    ----------
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    timeout : float, optional
        Fails when a connection takes longer than ``timeout`` seconds
        to initialize.

    load_operators : bool, optional
        Automatically load the math operators

    Examples
    --------
    Connect to an existing DPF server
    >>> from ansys.dpf import core as dpf
    >>> #server = dpf.connect_to_server(ip='127.0.0.1', port = 50054, as_global=False)
    >>> #base = dpf.BaseService(server=server)

    """

    def __init__(self, server=None, load_operators=True, timeout=5):
        """Initialize base service"""

        if server is None:
            server = serverlib._global_server()

        self._server = weakref.ref(server)
        self._stub = self._connect(timeout)

    def _connect(self, timeout=5):
        """Connect to dpf service within a given timeout"""
        stub = base_pb2_grpc.BaseServiceStub(self._server().channel)

        # verify connected
        if timeout is not None:
            state = grpc.channel_ready_future(self._server().channel)
            tstart = time.time()
            while (time.time() - tstart) < timeout and not state._matured:
                time.sleep(0.01)

            if not state._matured:
                raise IOError(
                    f"Unable to connect to DPF instance at {self._server()._input_ip} "
                    f"{self._server()._input_port}"
                )

        return stub

    def make_tmp_dir_server(self):
        """Create a temporary folder server side. Only one temporary folder can be created
        by server instance.
        The folder will be deleted when the server is stopped.

        Returns
        -------
        path : str
            path to the temporary dir
        """
        request = base_pb2.Empty()
        return self._stub.CreateTmpDir(request).server_file_path

    def load_library(self, filename, name="", symbol="LoadOperators"):
        """Dynamically load an operators library for dpf.core.
        Code containing this library's operators is generated in
        ansys.dpf.core.operators

        Parameters
        ----------
        filename : str
            Filename of the operator library.

        name : str, optional
            Library name.  Probably optional

        Examples
        --------
        Load the mesh operators for Windows (for Linux, just use
        'libmeshOperatorsCore.so' instead of 'meshOperatorsCore.dll')

        >>> from ansys.dpf import core as dpf
        >>> base = dpf.BaseService()
        >>> # base.load_library('meshOperatorsCore.dll', 'mesh_operators')

        """
        request = base_pb2.PluginRequest()
        request.name = name
        request.dllPath = filename
        request.symbol = symbol
        try:
            self._stub.Load(request)
        except Exception as e:
            raise IOError(
                f'Unable to load library "{filename}". File may not exist or'
                f" is missing dependencies:\n{str(e)}"
            )

        # TODO: fix code generation upload posix
        import os

        if os.name != "posix":
            local_dir = os.path.dirname(os.path.abspath(__file__))
            LOCAL_PATH = os.path.join(local_dir, "operators")

            # send local generated code
            TARGET_PATH = self.make_tmp_dir_server()
            self.upload_files_in_folder(TARGET_PATH, LOCAL_PATH, "py")

            # generate code
            from ansys.dpf.core.dpf_operator import Operator

            code_gen = Operator("python_generator")
            code_gen.connect(1, TARGET_PATH)
            code_gen.connect(0, filename)
            code_gen.connect(2, False)
            code_gen.run()

            self.download_files_in_folder(TARGET_PATH, LOCAL_PATH, "py")

    @property
    def server_info(self):
        """Send the request for server information and keep
        the info into a dictionary

        Returns
        -------
        info : dictionary
            dictionary with "server_ip", "server_port", "server_process_id"
            "server_version" keys
        """
        request = base_pb2.ServerInfoRequest()
        try:
            response = self._stub.GetServerInfo(request)
        except Exception as e:
            raise IOError(f"Unable to recover information from the server:\n{str(e)}")
        out = {
            "server_ip": response.ip,
            "server_port": response.port,
            "server_process_id": response.processId,
            "server_version": str(response.majorVersion)
            + "."
            + str(response.minorVersion),
        }
        return out

    def _description(self, dpf_entity_message):
        """Ask the server to describe the entity in input

        Parameters
        ----------
        dpf_entity_message : core.Operator._message, core.Workflow._message,
                             core.Scoping._message, core.Field._message,
        core.FieldContainer._message, core.MeshedRegion._message...

        Returns
        -------
           description : str
        """
        request = base_pb2.DescribeRequest()
        request.dpf_type_id = dpf_entity_message.id
        return self._stub.Describe(request).description

    def _get_separator(self, path):
        s1 = len(path.split("\\"))
        s2 = len(path.split("/"))
        if s2 > s1:
            # Linux case
            separator = "/"
        elif s1 > s2:
            # Windows case
            separator = "\\"
        return separator

    @protect_grpc
    def download_file(self, server_file_path, to_client_file_path):
        """Download a file from the server to the target client file path

        Notes
        -----
        Print a progress bar

        Parameters
        ----------
        server_file_path : str
            file path to dowload on the server side

        to_client_file_path: str
            file path target where the file will be located client side
        """
        request = base_pb2.DownloadFileRequest()
        request.server_file_path = server_file_path
        chunks = self._stub.DownloadFile(request)
        bar = _common_progress_bar("Downloading...", unit="KB")
        bar.start()
        i = 0
        with open(to_client_file_path, "wb") as f:
            for chunk in chunks:
                f.write(chunk.data.data)
                i += len(chunk.data.data) * 1e-3
                bar.update(i)
        bar.finish()

    @protect_grpc
    def download_files_in_folder(
        self, server_folder_path, to_client_folder_path, specific_extension=None
    ):
        """Download all the files from a folder of the server
        to the target client folder path

        Parameters
        ----------
        server_folder_path : str
            folder path to download on the server side

        to_client_folder_path: str
            folder path target where the files will be located client side

        specific_extension (optional) : str
            copies only the files with the given extension

        Notes
        -----
        Print a progress bar


        Returns
        -------
        paths : list of str
            new file paths client side
        """
        request = base_pb2.DownloadFileRequest()
        request.server_file_path = server_folder_path
        chunks = self._stub.DownloadFile(request)

        num_files = 1
        if chunks.initial_metadata()[0].key == "num_files":
            num_files = int(chunks.initial_metadata()[0].value)

        bar = _common_progress_bar("Downloading...", unit="files", tot_size=num_files)
        bar.start()

        server_path = ""

        import ntpath

        client_paths = []
        f = None
        for chunk in chunks:
            if chunk.data.server_file_path != server_path:
                server_path = chunk.data.server_file_path
                if (
                    specific_extension == None
                    or pathlib.Path(server_path).suffix == "." + specific_extension
                ):
                    separator = self._get_separator(server_path)
                    server_subpath = server_path.replace(
                        server_folder_path + separator, ""
                    )
                    subdir = ""
                    split = server_subpath.split(separator)
                    n = len(split)
                    i = 0
                    to_client_folder_path_copy = to_client_folder_path
                    if n > 1:
                        while i < (n - 1):
                            subdir = split[i]
                            subdir_path = os.path.join(
                                to_client_folder_path_copy, subdir
                            )
                            if not os.path.exists(subdir_path):
                                os.mkdir(subdir_path)
                            to_client_folder_path_copy = subdir_path
                            i += 1
                    cient_path = os.path.join(
                        to_client_folder_path_copy, ntpath.basename(server_path)
                    )
                    client_paths.append(cient_path)
                    f = open(cient_path, "wb")
                    try:
                        bar.update(len(client_paths))
                    except:
                        pass
                else:
                    f = None
            if f is not None:
                f.write(chunk.data.data)
        try:
            bar.finish()
        except:
            pass
        return client_paths

    @protect_grpc
    def upload_files_in_folder(
        self, to_server_folder_path, client_folder_path, specific_extension=None
    ):
        """Upload all the files from a folder of the client
        to the target server folder path.

        Parameters
        ----------
        to_server_folder_path : str
            folder path target where will be uploaded files on the server side

        client_folder_path: str
            folder path where the files that must be uploaded are located
            on client side

        specific_extension (optional) : str
            copies only the files with the given extension

        Returns
        -------
        paths : list of str
            new file paths server side
        """
        server_paths = []
        for root, subdirectories, files in os.walk(client_folder_path):
            for subdirectory in subdirectories:
                subdir = os.path.join(root, subdirectory)
                for filename in os.listdir(subdir):
                    f = os.path.join(subdir, filename)
                    server_paths = self._upload_and_get_server_path(
                        specific_extension,
                        f,
                        filename,
                        server_paths,
                        to_server_folder_path,
                        subdirectory,
                    )
            for file in files:
                f = os.path.join(root, file)
                server_paths = self._upload_and_get_server_path(
                    specific_extension, f, file, server_paths, to_server_folder_path
                )
            break
        return server_paths

    def _upload_and_get_server_path(
        self,
        specific_extension,
        f,
        filename,
        server_paths,
        to_server_folder_path,
        subdirectory=None,
    ):
        separator = self._get_separator(to_server_folder_path)

        if subdirectory is not None:
            to_server_file_path = (
                to_server_folder_path + separator + subdirectory + separator + filename
            )
        else:
            to_server_file_path = to_server_folder_path + separator + filename
        if ((specific_extension is not None) and (f.endswith(specific_extension))) or (
            specific_extension is None
        ):
            server_path = self._stub.UploadFile(
                self.__file_chunk_yielder(
                    file_path=f, to_server_file_path=to_server_file_path
                )
            ).server_file_path
            server_paths.append(server_path)
        return server_paths

    @protect_grpc
    def upload_file(self, file_path, to_server_file_path):
        """Upload a file from the client to the target server file path

        Parameters
        ----------
        file_path : str
            file path on the client side to upload

        to_server_file_path: str
            file path target where the file will be located server side

        Notes
        -----
        Print a progress bar

        Returns
        -------
           server_file_path : str
               path generated server side
        """
        if os.stat(file_path).st_size == 0:
            raise ValueError(file_path + " is empty")
        return self._stub.UploadFile(
            self.__file_chunk_yielder(file_path, to_server_file_path)
        ).server_file_path

    @protect_grpc
    def upload_file_in_tmp_folder(self, file_path, new_file_name=None):
        """Upload a file from the client to the server in a temporary folder
        deleted when the server is shutdown

        Parameters
        ----------
        file_path : str
            file path on the client side to upload

        new_file_name : str, optional
            name to give to the file server side,
            if no name is specified, the same name as the input file is given

        Notes
        -----
        Print a progress bar

        Returns
        -------
           server_file_path : str
               path generated server side
        """
        if new_file_name:
            file_name = new_file_name
        else:
            file_name = os.path.basename(file_path)
        if os.stat(file_path).st_size == 0:
            raise ValueError(file_path + " is empty")
        return self._stub.UploadFile(
            self.__file_chunk_yielder(
                file_path=file_path, to_server_file_path=file_name, use_tmp_dir=True
            )
        ).server_file_path

    def _prepare_shutdown(self):
        self._stub.PrepareShutdown(base_pb2.Empty())

    def __file_chunk_yielder(self, file_path, to_server_file_path, use_tmp_dir=False):
        request = base_pb2.UploadFileRequest()
        request.server_file_path = to_server_file_path
        request.use_temp_dir = use_tmp_dir

        tot_size = os.path.getsize(file_path) * 1e-3

        need_progress_bar = tot_size > 10000
        if need_progress_bar:
            bar = _common_progress_bar("Uploading...", "KB", tot_size)
            bar.start()
        i = 0
        with open(file_path, "rb") as f:
            while True:
                piece = f.read(DEFAULT_FILE_CHUNK_SIZE)
                if len(piece) == 0:
                    break
                request.data.data = piece
                yield request
                i += len(piece) * 1e-3
                if need_progress_bar:
                    try:
                        bar.update(min(i, tot_size))
                    except:
                        pass

        if need_progress_bar:
            try:
                bar.finish()
            except:
                pass



"""
MeshedRegion
============
"""
from ansys import dpf
from ansys.dpf.core import scoping
from ansys.dpf.core.check_version import server_meet_version
from ansys.dpf.core.common import locations, types
from ansys.dpf.core.elements import Elements, element_types
from ansys.dpf.core.nodes import Nodes
from ansys.dpf.core.plotter import Plotter as _DpfPlotter
from ansys.grpc.dpf import meshed_region_pb2, meshed_region_pb2_grpc


class MeshedRegion:
    """Represents a mesh from DPF.

    Parameters
    ----------
    num_nodes : int, optional
        Number of nodes to reserve for mesh creation. The default is ``None``.
    num_elements : int, optional
        Number of elements to reserve for mesh creation. The default is ``None``.
    mesh : ansys.grpc.dpf.meshed_region_pb2.MeshedRegion
        The default is ``None``.
    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Attributes
    ----------
    nodes : Nodes
        Entity containing all nodal properties.

    elements : Elements
        Entity containing all elemental properties.

    Examples
    --------
    Extract a meshed region from a model.

    >>> import ansys.dpf.core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.static_rst)
    >>> meshed_region = model.metadata.meshed_region

    Create a meshed region from scratch (line with 3 beam elements).

    >>> import ansys.dpf.core as dpf
    >>> meshed_region = dpf.MeshedRegion(num_nodes=4,num_elements=3)
    >>> i=0
    >>> for node in meshed_region.nodes.add_nodes(4):
    ...     node.id = i+1
    ...     node.coordinates = [float(i), float(i), 0.0]
    ...     i=i+1
    >>> i=0
    >>> for element in meshed_region.elements.add_elements(3):
    ...     element.id=i+1
    ...     element.connectivity = [i, i+1]
    ...     element.is_beam=True #or is_solid, is_beam, is_point
    ...     i=i+1
    >>> meshed_region.elements.add_beam_element(id=4,connectivity=[3,0])

    """

    def __init__(self, num_nodes=None, num_elements=None, mesh=None, server=None):

        if server is None:
            server = dpf.core._global_server()

        self._server = server
        self._stub = self._connect()

        if isinstance(mesh, MeshedRegion):
            self._message = mesh._mesh
        elif isinstance(mesh, meshed_region_pb2.MeshedRegion):
            self._message = mesh
        elif mesh == None:
            self.__send_init_request(num_nodes, num_elements)
        else:  # support_pb2.Support
            self._message = meshed_region_pb2.MeshedRegion()
            self._message.id = mesh.id

        self._full_grid = None
        self._elements = None
        self._nodes = None

    def _get_scoping(self, loc=locations.nodal):
        """
        Parameters
        ----------
        loc : str or ansys.dpf.core.common.locations, optional
            location of the requested scoping ("Nodal", "Elemental"...)

        Returns
        -------
        scoping : Scoping
            ids of the elements or nodes of the mesh
        """
        request = meshed_region_pb2.GetScopingRequest(mesh=self._message)
        request.loc.location = loc
        out = self._stub.GetScoping(request)
        return scoping.Scoping(scoping=out, server=self._server)

    @property
    def elements(self):
        """All elemental properties of the mesh, such as connectivity and element types.

        Returns
        -------
        elements : Elements
            Elements belonging to the meshed region.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> meshed_region = model.metadata.meshed_region
        >>> elements = meshed_region.elements
        >>> print(elements)
        DPF Elements object with 8 elements

        """
        if self._elements is None:
            self._elements = Elements(self)
        return self._elements

    @property
    def nodes(self):
        """All nodal properties of the mesh, such as node coordinates and nodal connectivity.

        Returns
        -------
        nodes : Nodes
            Nodes belonging to the meshed region

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> meshed_region = model.metadata.meshed_region
        >>> nodes = meshed_region.nodes
        >>> nodes.n_nodes
        81

        """
        if self._nodes is None:
            self._nodes = Nodes(self)
        return self._nodes

    @property
    def unit(self):
        """Unit of the meshed region.

        This unit is the same as the unit of the coordinates of the meshed region.

        Returns
        -------
        unit : str
        """
        return self._get_unit()

    @unit.setter
    def unit(self, value):
        """Unit type.

        Parameters
        ----------
        unit : str
        """
        return self._set_unit(value)

    def _get_unit(self):
        """Retrieve the unit type.

        Returns
        -------
        unit : str
        """
        return self._stub.List(self._message).unit

    def _set_unit(self, unit):
        """Set the unit of the meshed region.

        Parameters
        ----------
        unit: str
        """
        request = meshed_region_pb2.UpdateMeshedRegionRequest()
        request.meshed_region.CopyFrom(self._message)
        request.unit = unit
        return self._stub.UpdateRequest(request)

    def __del__(self):
        try:
            self._stub.Delete(self._message)
        except:
            pass

    def _connect(self):
        """Connect to the gRPC service containing the reader."""
        return meshed_region_pb2_grpc.MeshedRegionServiceStub(self._server.channel)

    def __str__(self):
        from ansys.dpf.core.core import _description

        return _description(self._message, self._server)

    @property
    def available_named_selections(self):
        """List of available named selections.

        Returns
        -------
        named_selections : list str
        """
        return self._stub.List(self._message).named_selections

    def named_selection(self, named_selection):
        """Scoping containing the list of nodes or elements in the named selection.

        Parameters
        ----------
        named_selection : str
            Name of the named selection.

        Returns
        -------
        named_selection : Scoping
        """
        if server_meet_version("2.1", self._server):
            request = meshed_region_pb2.GetScopingRequest(mesh=self._message)
            request.named_selection = named_selection
            out = self._stub.GetScoping(request)
            return scoping.Scoping(scoping=out, server=self._server)
        else:
            if hasattr(self, "_stream_provider"):
                from ansys.dpf.core.dpf_operator import Operator

                op = Operator("scoping_provider_by_ns", server=self._server)
                op.connect(1, named_selection)
                op.connect(3, self._stream_provider, 0)
                return op.get_output(0, types.scoping)
            else:
                raise Exception(
                    "Getting a named selection from a meshed region is "
                    "only implemented for meshed region created from a "
                    "model for server version 2.0. Please update your server."
                )

    def _set_stream_provider(self, stream_provider):
        self._stream_provider = stream_provider

    # NOTE: kept only for reference as the mesh operator is being moved out of dpf
    # def write_vtk(self, filename, skin_only=True):
    #     """Return a vtk mesh"""
    #     # filename = os.path.join(tempfile.gettempdir(),
    #                             # '%s.vtk' % next(tempfile._get_candidate_names()))

    #     vtk_exp = self._model.operator("vtk_export")
    #     vtk_exp.connect(0, filename)

    #     mesh = self._model.operator("mapdl::rst::MeshProvider")
    #     mesh.connect(4, self._model.data_sources)

    #     if skin_only:
    #         skin = self._model.operator("meshed_skin_sector")
    #         skin.connect(0, mesh, 0)
    #         vtk_exp.connect(1, skin, 0)
    #     else:
    #         vtk_exp.connect(1, mesh, 0)

    #     vtk_exp.run()
    #     if not os.path.isfile(filename):
    #         raise FileNotFoundError('VTK mesh not written to disk')

    # @property
    # def skin(self):
    #     """Surface of the meshed region."""
    #     mesh = self._model.operator("mapdl::rst::MeshProvider")
    #     mesh.connect(4, self._model.data_sources)

    #     skin = self._model.operator("meshed_skin_sector")
    #     skin.connect(0, mesh, 0)
    #     # skin.connect(4, self)

    #     skin.get_output(0, types.meshed_region)

    #     name = None
    #     if self._name:
    #         name = 'Skin of %s' % self._name
    #     self._message = skin.get_output(0, types.meshed_region)
    #     return MeshedRegion(self._server.channel, skin, self._model, name)

    def _as_vtk(self, as_linear=True, include_ids=False):
        """Convert DPF mesh to a PyVista unstructured grid."""
        nodes = self.nodes.coordinates_field.data
        etypes = self.elements.element_types_field.data
        conn = self.elements.connectivities_field.data
        try:
            from ansys.dpf.core.vtk_helper import dpf_mesh_to_vtk
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "To use plotting capabilities, please install pyvista "
                "with :\n pip install pyvista>=0.24.0"
            )

        grid = dpf_mesh_to_vtk(nodes, etypes, conn, as_linear)

        # consider adding this when scoping request is faster
        if include_ids:
            grid["node_ids"] = self.nodes.scoping.ids
            grid["element_ids"] = self.elements.scoping.ids

        return grid

    @property
    def grid(self):
        """Unstructured grid in VTK fromat from PyVista.

        Returns
        -------
        pyvista.UnstructuredGrid
            UnstructuredGrid of the mesh.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> meshed_region = model.metadata.meshed_region
        >>> grid = meshed_region.grid

        Plot this grid directly.

        >>> grid.plot()

        Extract the surface mesh of this grid

        >>> mesh = grid.extract_surface()

        """
        if self._full_grid is None:
            self._full_grid = self._as_vtk()
        return self._full_grid

    def plot(
        self,
        field_or_fields_container=None,
        notebook=None,
        shell_layers=None,
        off_screen=None,
        show_axes=True,
        **kwargs
    ):
        """Plot the field or fields container on the mesh.

        Parameters
        ----------
        field_or_fields_container : dpf.core.Field or dpf.core.FieldsContainer
            Field or fields container to plot. The default is ``None``.
        notebook : bool, optional
            Whether the plotting in the notebook is 2D or 3D. The default is
            ``None``, in which case the plotting is 2D.
        shell_layers : core.shell_layers, optional
            Enum used to set the shell layers if the model to plot contains shell elements.
        off_screen : bool, optional
            Whether to render the plot off screen, which is useful for automated screenshots.
            The default is "None", in which case the plot renders off screen.
        show_axes : bool, optional
            Whether to show a VTK axes widget. The default is ``True``.
        **kwargs : optional
            Additional keyword arguments for the plotter. For additional keyword
            arguments, see ``help(pyvista.plot)``.

        Examples
        --------
        Plot the displacement field from an example file.

        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> disp = model.results.displacement()
        >>> field = disp.outputs.fields_container()[0]
        >>> model.metadata.meshed_region.plot(field)

        """
        pl = _DpfPlotter(self)
        if field_or_fields_container is not None:
            return pl.plot_contour(
                field_or_fields_container,
                notebook,
                shell_layers,
                off_screen,
                show_axes,
                **kwargs
            )

        # otherwise, simply plot self
        kwargs["notebook"] = notebook
        return pl.plot_mesh(**kwargs)

    def deep_copy(self, server=None):
        """Create a deep copy of the meshed region's data on a given server.

        This method is useful for passing data from one server instance to another.

        .. warning::
           Only nodes scoping and coordinates and elements scoping, connectivity,
           and types are copied. The eventual property field for elemental properties
           and named selection will not be copied.

        Parameters
        ----------
        server : ansys.dpf.core.server, optional
            Server with the channel connected to the remote or local instance.
            The default is ``None``, in which case an attempt is made to use the
            global server.

        Returns
        -------
        mesh_copy : MeshedRegion

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> meshed_region = model.metadata.meshed_region
        >>> other_server = dpf.start_local_server(as_global=False)
        >>> deep_copy = meshed_region.deep_copy(server=other_server)

        """
        node_ids = self.nodes.scoping.ids
        element_ids = self.elements.scoping.ids
        mesh = MeshedRegion(
            num_nodes=len(node_ids), num_elements=len(element_ids), server=server
        )
        with self.nodes.coordinates_field.as_local_field() as coord:
            for i, node in enumerate(mesh.nodes.add_nodes(len(node_ids))):
                node.id = node_ids[i]
                node.coordinates = coord.get_entity_data(i)
        with self.elements.connectivities_field.as_local_field() as connect:
            with self.elements.element_types_field.as_local_field() as types:
                for i, elem in enumerate(mesh.elements.add_elements(len(element_ids))):
                    elem.id = element_ids[i]
                    elem.connectivity = connect.get_entity_data(i)
                    elem.shape = element_types.shape(types.get_entity_data(i)[0])
        mesh.unit = self.unit
        return mesh

    def __send_init_request(self, num_nodes=0, num_elements=0):
        request = meshed_region_pb2.CreateRequest()
        if num_nodes:
            request.num_nodes_reserved = num_nodes
        if num_elements:
            request.num_elements_reserved = num_elements
        self._message = self._stub.Create(request)



"""
poynting_vector
===============
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.FEMutils plugin, from "result" category
"""

class poynting_vector(Operator):
    """Compute the Poynting Vector

      available inputs:
        - fields_containerA (FieldsContainer)
        - fields_containerB (FieldsContainer)
        - fields_containerC (FieldsContainer)
        - fields_containerD (FieldsContainer)
        - meshed_region (MeshedRegion) (optional)
        - int32 (int) (optional)

      available outputs:
        - fields_container (FieldsContainer)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.result.poynting_vector()

      >>> # Make input connections
      >>> my_fields_containerA = dpf.FieldsContainer()
      >>> op.inputs.fields_containerA.connect(my_fields_containerA)
      >>> my_fields_containerB = dpf.FieldsContainer()
      >>> op.inputs.fields_containerB.connect(my_fields_containerB)
      >>> my_fields_containerC = dpf.FieldsContainer()
      >>> op.inputs.fields_containerC.connect(my_fields_containerC)
      >>> my_fields_containerD = dpf.FieldsContainer()
      >>> op.inputs.fields_containerD.connect(my_fields_containerD)
      >>> my_meshed_region = dpf.MeshedRegion()
      >>> op.inputs.meshed_region.connect(my_meshed_region)
      >>> my_int32 = int()
      >>> op.inputs.int32.connect(my_int32)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.result.poynting_vector(fields_containerA=my_fields_containerA,fields_containerB=my_fields_containerB,fields_containerC=my_fields_containerC,fields_containerD=my_fields_containerD,meshed_region=my_meshed_region,int32=my_int32)

      >>> # Get output data
      >>> result_fields_container = op.outputs.fields_container()"""
    def __init__(self, fields_containerA=None, fields_containerB=None, fields_containerC=None, fields_containerD=None, meshed_region=None, int32=None, config=None, server=None):
        super().__init__(name="PoyntingVector", config = config, server = server)
        self._inputs = InputsPoyntingVector(self)
        self._outputs = OutputsPoyntingVector(self)
        if fields_containerA !=None:
            self.inputs.fields_containerA.connect(fields_containerA)
        if fields_containerB !=None:
            self.inputs.fields_containerB.connect(fields_containerB)
        if fields_containerC !=None:
            self.inputs.fields_containerC.connect(fields_containerC)
        if fields_containerD !=None:
            self.inputs.fields_containerD.connect(fields_containerD)
        if meshed_region !=None:
            self.inputs.meshed_region.connect(meshed_region)
        if int32 !=None:
            self.inputs.int32.connect(int32)

    @staticmethod
    def _spec():
        spec = Specification(description="""Compute the Poynting Vector""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "fields_containerA", type_names=["fields_container"], optional=False, document=""""""), 
                                 1 : PinSpecification(name = "fields_containerB", type_names=["fields_container"], optional=False, document=""""""), 
                                 2 : PinSpecification(name = "fields_containerC", type_names=["fields_container"], optional=False, document=""""""), 
                                 3 : PinSpecification(name = "fields_containerD", type_names=["fields_container"], optional=False, document=""""""), 
                                 4 : PinSpecification(name = "meshed_region", type_names=["abstract_meshed_region"], optional=True, document="""the mesh region in this pin have to be boundary or skin mesh"""), 
                                 5 : PinSpecification(name = "int32", type_names=["int32"], optional=True, document="""load step number, if it's specified, the Poynting Vector is computed only on the substeps of this step""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "PoyntingVector")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsPoyntingVector 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsPoyntingVector 
        """
        return super().outputs


#internal name: PoyntingVector
#scripting name: poynting_vector
class InputsPoyntingVector(_Inputs):
    """Intermediate class used to connect user inputs to poynting_vector operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.result.poynting_vector()
      >>> my_fields_containerA = dpf.FieldsContainer()
      >>> op.inputs.fields_containerA.connect(my_fields_containerA)
      >>> my_fields_containerB = dpf.FieldsContainer()
      >>> op.inputs.fields_containerB.connect(my_fields_containerB)
      >>> my_fields_containerC = dpf.FieldsContainer()
      >>> op.inputs.fields_containerC.connect(my_fields_containerC)
      >>> my_fields_containerD = dpf.FieldsContainer()
      >>> op.inputs.fields_containerD.connect(my_fields_containerD)
      >>> my_meshed_region = dpf.MeshedRegion()
      >>> op.inputs.meshed_region.connect(my_meshed_region)
      >>> my_int32 = int()
      >>> op.inputs.int32.connect(my_int32)
    """
    def __init__(self, op: Operator):
        super().__init__(poynting_vector._spec().inputs, op)
        self._fields_containerA = Input(poynting_vector._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._fields_containerA)
        self._fields_containerB = Input(poynting_vector._spec().input_pin(1), 1, op, -1) 
        self._inputs.append(self._fields_containerB)
        self._fields_containerC = Input(poynting_vector._spec().input_pin(2), 2, op, -1) 
        self._inputs.append(self._fields_containerC)
        self._fields_containerD = Input(poynting_vector._spec().input_pin(3), 3, op, -1) 
        self._inputs.append(self._fields_containerD)
        self._meshed_region = Input(poynting_vector._spec().input_pin(4), 4, op, -1) 
        self._inputs.append(self._meshed_region)
        self._int32 = Input(poynting_vector._spec().input_pin(5), 5, op, -1) 
        self._inputs.append(self._int32)

    @property
    def fields_containerA(self):
        """Allows to connect fields_containerA input to the operator

        Parameters
        ----------
        my_fields_containerA : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.poynting_vector()
        >>> op.inputs.fields_containerA.connect(my_fields_containerA)
        >>> #or
        >>> op.inputs.fields_containerA(my_fields_containerA)

        """
        return self._fields_containerA

    @property
    def fields_containerB(self):
        """Allows to connect fields_containerB input to the operator

        Parameters
        ----------
        my_fields_containerB : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.poynting_vector()
        >>> op.inputs.fields_containerB.connect(my_fields_containerB)
        >>> #or
        >>> op.inputs.fields_containerB(my_fields_containerB)

        """
        return self._fields_containerB

    @property
    def fields_containerC(self):
        """Allows to connect fields_containerC input to the operator

        Parameters
        ----------
        my_fields_containerC : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.poynting_vector()
        >>> op.inputs.fields_containerC.connect(my_fields_containerC)
        >>> #or
        >>> op.inputs.fields_containerC(my_fields_containerC)

        """
        return self._fields_containerC

    @property
    def fields_containerD(self):
        """Allows to connect fields_containerD input to the operator

        Parameters
        ----------
        my_fields_containerD : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.poynting_vector()
        >>> op.inputs.fields_containerD.connect(my_fields_containerD)
        >>> #or
        >>> op.inputs.fields_containerD(my_fields_containerD)

        """
        return self._fields_containerD

    @property
    def meshed_region(self):
        """Allows to connect meshed_region input to the operator

        - pindoc: the mesh region in this pin have to be boundary or skin mesh

        Parameters
        ----------
        my_meshed_region : MeshedRegion, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.poynting_vector()
        >>> op.inputs.meshed_region.connect(my_meshed_region)
        >>> #or
        >>> op.inputs.meshed_region(my_meshed_region)

        """
        return self._meshed_region

    @property
    def int32(self):
        """Allows to connect int32 input to the operator

        - pindoc: load step number, if it's specified, the Poynting Vector is computed only on the substeps of this step

        Parameters
        ----------
        my_int32 : int, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.poynting_vector()
        >>> op.inputs.int32.connect(my_int32)
        >>> #or
        >>> op.inputs.int32(my_int32)

        """
        return self._int32

class OutputsPoyntingVector(_Outputs):
    """Intermediate class used to get outputs from poynting_vector operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.result.poynting_vector()
      >>> # Connect inputs : op.inputs. ...
      >>> result_fields_container = op.outputs.fields_container()
    """
    def __init__(self, op: Operator):
        super().__init__(poynting_vector._spec().outputs, op)
        self._fields_container = Output(poynting_vector._spec().output_pin(0), 0, op) 
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator


        Returns
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.poynting_vector()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container() 
        """
        return self._fields_container




"""
min_max_over_label_fc
=====================
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.Native plugin, from "min_max" category
"""

class min_max_over_label_fc(Operator):
    """Create two fields (0 min 1 max) by looping over the fields container in input and taking the min/max value by component through all the fields having the same id for the label set in input (in pin 1). If no label is specified or if the specified label doesn't exist, the operation is done over all the fields. The fields out are located on the label set in input, so their scoping are the labels ids.For each min max value, the label id for one other fields container labels is kept and returned in a scoping in pin 2 (min) and 3 (max).The field's scoping ids of the value kept in min max are also returned in the scopings in pin 4 (min) and 5 (max).

      available inputs:
        - fields_container (FieldsContainer)
        - label (str)

      available outputs:
        - field_min (Field)
        - field_max (Field)
        - domain_ids_min (Scoping)
        - domain_ids_max (Scoping)
        - scoping_ids_min (Scoping)
        - scoping_ids_max (Scoping)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.min_max.min_max_over_label_fc()

      >>> # Make input connections
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)
      >>> my_label = str()
      >>> op.inputs.label.connect(my_label)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.min_max.min_max_over_label_fc(fields_container=my_fields_container,label=my_label)

      >>> # Get output data
      >>> result_field_min = op.outputs.field_min()
      >>> result_field_max = op.outputs.field_max()
      >>> result_domain_ids_min = op.outputs.domain_ids_min()
      >>> result_domain_ids_max = op.outputs.domain_ids_max()
      >>> result_scoping_ids_min = op.outputs.scoping_ids_min()
      >>> result_scoping_ids_max = op.outputs.scoping_ids_max()"""
    def __init__(self, fields_container=None, label=None, config=None, server=None):
        super().__init__(name="min_max_over_label_fc", config = config, server = server)
        self._inputs = InputsMinMaxOverLabelFc(self)
        self._outputs = OutputsMinMaxOverLabelFc(self)
        if fields_container !=None:
            self.inputs.fields_container.connect(fields_container)
        if label !=None:
            self.inputs.label.connect(label)

    @staticmethod
    def _spec():
        spec = Specification(description="""Create two fields (0 min 1 max) by looping over the fields container in input and taking the min/max value by component through all the fields having the same id for the label set in input (in pin 1). If no label is specified or if the specified label doesn't exist, the operation is done over all the fields. The fields out are located on the label set in input, so their scoping are the labels ids.For each min max value, the label id for one other fields container labels is kept and returned in a scoping in pin 2 (min) and 3 (max).The field's scoping ids of the value kept in min max are also returned in the scopings in pin 4 (min) and 5 (max).""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document=""""""), 
                                 1 : PinSpecification(name = "label", type_names=["string"], optional=False, document="""label name from the fields container""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "field_min", type_names=["field"], optional=False, document=""""""), 
                                 1 : PinSpecification(name = "field_max", type_names=["field"], optional=False, document=""""""), 
                                 2 : PinSpecification(name = "domain_ids_min", type_names=["scoping"], optional=True, document=""""""), 
                                 3 : PinSpecification(name = "domain_ids_max", type_names=["scoping"], optional=True, document=""""""), 
                                 4 : PinSpecification(name = "scoping_ids_min", type_names=["scoping"], optional=False, document=""""""), 
                                 5 : PinSpecification(name = "scoping_ids_max", type_names=["scoping"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "min_max_over_label_fc")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsMinMaxOverLabelFc 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsMinMaxOverLabelFc 
        """
        return super().outputs


#internal name: min_max_over_label_fc
#scripting name: min_max_over_label_fc
class InputsMinMaxOverLabelFc(_Inputs):
    """Intermediate class used to connect user inputs to min_max_over_label_fc operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.min_max.min_max_over_label_fc()
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)
      >>> my_label = str()
      >>> op.inputs.label.connect(my_label)
    """
    def __init__(self, op: Operator):
        super().__init__(min_max_over_label_fc._spec().inputs, op)
        self._fields_container = Input(min_max_over_label_fc._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._fields_container)
        self._label = Input(min_max_over_label_fc._spec().input_pin(1), 1, op, -1) 
        self._inputs.append(self._label)

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator

        Parameters
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.min_max.min_max_over_label_fc()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> #or
        >>> op.inputs.fields_container(my_fields_container)

        """
        return self._fields_container

    @property
    def label(self):
        """Allows to connect label input to the operator

        - pindoc: label name from the fields container

        Parameters
        ----------
        my_label : str, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.min_max.min_max_over_label_fc()
        >>> op.inputs.label.connect(my_label)
        >>> #or
        >>> op.inputs.label(my_label)

        """
        return self._label

class OutputsMinMaxOverLabelFc(_Outputs):
    """Intermediate class used to get outputs from min_max_over_label_fc operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.min_max.min_max_over_label_fc()
      >>> # Connect inputs : op.inputs. ...
      >>> result_field_min = op.outputs.field_min()
      >>> result_field_max = op.outputs.field_max()
      >>> result_domain_ids_min = op.outputs.domain_ids_min()
      >>> result_domain_ids_max = op.outputs.domain_ids_max()
      >>> result_scoping_ids_min = op.outputs.scoping_ids_min()
      >>> result_scoping_ids_max = op.outputs.scoping_ids_max()
    """
    def __init__(self, op: Operator):
        super().__init__(min_max_over_label_fc._spec().outputs, op)
        self._field_min = Output(min_max_over_label_fc._spec().output_pin(0), 0, op) 
        self._outputs.append(self._field_min)
        self._field_max = Output(min_max_over_label_fc._spec().output_pin(1), 1, op) 
        self._outputs.append(self._field_max)
        self._domain_ids_min = Output(min_max_over_label_fc._spec().output_pin(2), 2, op) 
        self._outputs.append(self._domain_ids_min)
        self._domain_ids_max = Output(min_max_over_label_fc._spec().output_pin(3), 3, op) 
        self._outputs.append(self._domain_ids_max)
        self._scoping_ids_min = Output(min_max_over_label_fc._spec().output_pin(4), 4, op) 
        self._outputs.append(self._scoping_ids_min)
        self._scoping_ids_max = Output(min_max_over_label_fc._spec().output_pin(5), 5, op) 
        self._outputs.append(self._scoping_ids_max)

    @property
    def field_min(self):
        """Allows to get field_min output of the operator


        Returns
        ----------
        my_field_min : Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.min_max.min_max_over_label_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field_min = op.outputs.field_min() 
        """
        return self._field_min

    @property
    def field_max(self):
        """Allows to get field_max output of the operator


        Returns
        ----------
        my_field_max : Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.min_max.min_max_over_label_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field_max = op.outputs.field_max() 
        """
        return self._field_max

    @property
    def domain_ids_min(self):
        """Allows to get domain_ids_min output of the operator


        Returns
        ----------
        my_domain_ids_min : Scoping, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.min_max.min_max_over_label_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_domain_ids_min = op.outputs.domain_ids_min() 
        """
        return self._domain_ids_min

    @property
    def domain_ids_max(self):
        """Allows to get domain_ids_max output of the operator


        Returns
        ----------
        my_domain_ids_max : Scoping, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.min_max.min_max_over_label_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_domain_ids_max = op.outputs.domain_ids_max() 
        """
        return self._domain_ids_max

    @property
    def scoping_ids_min(self):
        """Allows to get scoping_ids_min output of the operator


        Returns
        ----------
        my_scoping_ids_min : Scoping, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.min_max.min_max_over_label_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_scoping_ids_min = op.outputs.scoping_ids_min() 
        """
        return self._scoping_ids_min

    @property
    def scoping_ids_max(self):
        """Allows to get scoping_ids_max output of the operator


        Returns
        ----------
        my_scoping_ids_max : Scoping, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.min_max.min_max_over_label_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_scoping_ids_max = op.outputs.scoping_ids_max() 
        """
        return self._scoping_ids_max




"""
add_rigid_body_motion_fc
========================
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from mapdlOperatorsCore plugin, from "result" category
"""

class add_rigid_body_motion_fc(Operator):
    """Adds a given rigid translation, center and rotation from a displacement field. The rotation is given in terms of rotations angles. Note that the displacement field has to be in the global coordinate sytem

      available inputs:
        - fields_container (FieldsContainer)
        - translation_field (Field)
        - rotation_field (Field)
        - center_field (Field)
        - mesh (MeshedRegion) (optional)

      available outputs:
        - fields_container (FieldsContainer)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.result.add_rigid_body_motion_fc()

      >>> # Make input connections
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)
      >>> my_translation_field = dpf.Field()
      >>> op.inputs.translation_field.connect(my_translation_field)
      >>> my_rotation_field = dpf.Field()
      >>> op.inputs.rotation_field.connect(my_rotation_field)
      >>> my_center_field = dpf.Field()
      >>> op.inputs.center_field.connect(my_center_field)
      >>> my_mesh = dpf.MeshedRegion()
      >>> op.inputs.mesh.connect(my_mesh)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.result.add_rigid_body_motion_fc(fields_container=my_fields_container,translation_field=my_translation_field,rotation_field=my_rotation_field,center_field=my_center_field,mesh=my_mesh)

      >>> # Get output data
      >>> result_fields_container = op.outputs.fields_container()"""
    def __init__(self, fields_container=None, translation_field=None, rotation_field=None, center_field=None, mesh=None, config=None, server=None):
        super().__init__(name="RigidBodyAddition_fc", config = config, server = server)
        self._inputs = InputsAddRigidBodyMotionFc(self)
        self._outputs = OutputsAddRigidBodyMotionFc(self)
        if fields_container !=None:
            self.inputs.fields_container.connect(fields_container)
        if translation_field !=None:
            self.inputs.translation_field.connect(translation_field)
        if rotation_field !=None:
            self.inputs.rotation_field.connect(rotation_field)
        if center_field !=None:
            self.inputs.center_field.connect(center_field)
        if mesh !=None:
            self.inputs.mesh.connect(mesh)

    @staticmethod
    def _spec():
        spec = Specification(description="""Adds a given rigid translation, center and rotation from a displacement field. The rotation is given in terms of rotations angles. Note that the displacement field has to be in the global coordinate sytem""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document=""""""), 
                                 1 : PinSpecification(name = "translation_field", type_names=["field"], optional=False, document=""""""), 
                                 2 : PinSpecification(name = "rotation_field", type_names=["field"], optional=False, document=""""""), 
                                 3 : PinSpecification(name = "center_field", type_names=["field"], optional=False, document=""""""), 
                                 7 : PinSpecification(name = "mesh", type_names=["abstract_meshed_region"], optional=True, document="""default is the mesh in the support""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "RigidBodyAddition_fc")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsAddRigidBodyMotionFc 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsAddRigidBodyMotionFc 
        """
        return super().outputs


#internal name: RigidBodyAddition_fc
#scripting name: add_rigid_body_motion_fc
class InputsAddRigidBodyMotionFc(_Inputs):
    """Intermediate class used to connect user inputs to add_rigid_body_motion_fc operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.result.add_rigid_body_motion_fc()
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)
      >>> my_translation_field = dpf.Field()
      >>> op.inputs.translation_field.connect(my_translation_field)
      >>> my_rotation_field = dpf.Field()
      >>> op.inputs.rotation_field.connect(my_rotation_field)
      >>> my_center_field = dpf.Field()
      >>> op.inputs.center_field.connect(my_center_field)
      >>> my_mesh = dpf.MeshedRegion()
      >>> op.inputs.mesh.connect(my_mesh)
    """
    def __init__(self, op: Operator):
        super().__init__(add_rigid_body_motion_fc._spec().inputs, op)
        self._fields_container = Input(add_rigid_body_motion_fc._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._fields_container)
        self._translation_field = Input(add_rigid_body_motion_fc._spec().input_pin(1), 1, op, -1) 
        self._inputs.append(self._translation_field)
        self._rotation_field = Input(add_rigid_body_motion_fc._spec().input_pin(2), 2, op, -1) 
        self._inputs.append(self._rotation_field)
        self._center_field = Input(add_rigid_body_motion_fc._spec().input_pin(3), 3, op, -1) 
        self._inputs.append(self._center_field)
        self._mesh = Input(add_rigid_body_motion_fc._spec().input_pin(7), 7, op, -1) 
        self._inputs.append(self._mesh)

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator

        Parameters
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.add_rigid_body_motion_fc()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> #or
        >>> op.inputs.fields_container(my_fields_container)

        """
        return self._fields_container

    @property
    def translation_field(self):
        """Allows to connect translation_field input to the operator

        Parameters
        ----------
        my_translation_field : Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.add_rigid_body_motion_fc()
        >>> op.inputs.translation_field.connect(my_translation_field)
        >>> #or
        >>> op.inputs.translation_field(my_translation_field)

        """
        return self._translation_field

    @property
    def rotation_field(self):
        """Allows to connect rotation_field input to the operator

        Parameters
        ----------
        my_rotation_field : Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.add_rigid_body_motion_fc()
        >>> op.inputs.rotation_field.connect(my_rotation_field)
        >>> #or
        >>> op.inputs.rotation_field(my_rotation_field)

        """
        return self._rotation_field

    @property
    def center_field(self):
        """Allows to connect center_field input to the operator

        Parameters
        ----------
        my_center_field : Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.add_rigid_body_motion_fc()
        >>> op.inputs.center_field.connect(my_center_field)
        >>> #or
        >>> op.inputs.center_field(my_center_field)

        """
        return self._center_field

    @property
    def mesh(self):
        """Allows to connect mesh input to the operator

        - pindoc: default is the mesh in the support

        Parameters
        ----------
        my_mesh : MeshedRegion, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.add_rigid_body_motion_fc()
        >>> op.inputs.mesh.connect(my_mesh)
        >>> #or
        >>> op.inputs.mesh(my_mesh)

        """
        return self._mesh

class OutputsAddRigidBodyMotionFc(_Outputs):
    """Intermediate class used to get outputs from add_rigid_body_motion_fc operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.result.add_rigid_body_motion_fc()
      >>> # Connect inputs : op.inputs. ...
      >>> result_fields_container = op.outputs.fields_container()
    """
    def __init__(self, op: Operator):
        super().__init__(add_rigid_body_motion_fc._spec().outputs, op)
        self._fields_container = Output(add_rigid_body_motion_fc._spec().output_pin(0), 0, op) 
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator


        Returns
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.result.add_rigid_body_motion_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container() 
        """
        return self._fields_container

